# Standard library imports
import gc
import math
import os
import shutil
import sys
import warnings

# from collections import namedtuple
from contextlib import redirect_stdout
from time import time
from typing import Optional, Union, List, Dict, Tuple, Any, Callable

# Third-party imports
## For plotting
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.backends.backend_pdf import PdfPages

## For stats and numeric operations
import numpy as np
import pandas as pd
from scipy import stats

# scikit-learn imports
from sklearn.base import clone
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.impute import SimpleImputer
from sklearn.impute._base import _check_inputs_dtype

## For warnings
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._testing import ignore_warnings

## Required for IterativeImputer.fit_transform()
from sklearn.utils import check_random_state, _safe_indexing, is_scalar_nan
from sklearn.utils._mask import _get_mask
from sklearn.utils.validation import FLOAT_DTYPES

# Grid search imports
from sklearn.model_selection import RandomizedSearchCV, GridSearchCV
from sklearn.model_selection import StratifiedKFold

# Genetic algorithm grid search imports
from sklearn_genetic import GASearchCV
from sklearn_genetic.callbacks import ConsecutiveStopping, DeltaThreshold
from sklearn_genetic.plots import plot_fitness_evolution
from sklearn.preprocessing import (
    LabelEncoder,
    OneHotEncoder,
    OrdinalEncoder,
    TargetEncoder,
)
from sklearn.exceptions import NotFittedError
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.pipeline import make_pipeline

from xgboost import XGBClassifier

# Custom function imports
try:
    from .. import simple_imputers
    from ...utils.plotting import Plotting
    from ...utils.misc import get_processor_name
    from ...utils.misc import HiddenPrints
    from ...utils.misc import isnotebook
except (ModuleNotFoundError, ValueError, ImportError):
    from impute import simple_imputers
    from utils.plotting import Plotting
    from utils.misc import get_processor_name
    from utils.misc import HiddenPrints
    from utils.misc import isnotebook

# Uses scikit-learn-intellex package if CPU is Intel
if get_processor_name().strip().startswith("Intel"):
    try:
        from sklearnex import patch_sklearn

        patch_sklearn(verbose=False)
    except (ImportError, TypeError):
        print(
            "Processor not compatible with scikit-learn-intelex; using "
            "default configuration"
        )

is_notebook = isnotebook()

if is_notebook:
    from tqdm.notebook import tqdm as progressbar
else:
    if sys.platform == "linux" or sys.platform == "linux2":
        from tqdm.auto import tqdm as progressbar
    else:
        from tqdm import tqdm as progressbar

# NOTE: Removed ImputeTriplets to save memory.
# ImputerTriplet is there so that the missing values
# can be predicted on an already-fit model using just the
# transform method. I didn't need it, so I removed it
# because it was saving thousands of fit estimator models into the object

# _ImputerTripletGrid = namedtuple(
# 	'_ImputerTripletGrid', ['feat_idx', 'neighbor_feat_idx', 'estimator'])


class IterativeImputerGridSearch(IterativeImputer):
    """Overridden IterativeImputer methods.

    Herein, two types of grid searches (RandomizedSearchCV and GASearchCV), progress status updates, and several other improvements have been added. IterativeImputer is a multivariate imputer that estimates each feature from all the others. A strategy for imputing missing values by modeling each feature with missing values as a function of other features in a round-robin fashion.Read more in the scikit-learn User Guide for IterativeImputer. scikit-learn version added: 0.21. NOTE: This estimator is still **experimental** for now: the predictions and the API might change without any deprecation cycle. To use it, you need to explicitly import ``enable_iterative_imputer``\.

    IterativeImputer is based on the R MICE (Multivariate Imputation by Chained Equationspackage) [van Buuren & Groothuis-Oudshoorn, 2011]_. See [Buck, 1960]_ for more information about multiple versus single imputations.

    >>> # explicitly require this experimental feature
    >>> from sklearn.experimental import enable_iterative_imputer
    >>>
    >>> # now you can import normally from sklearn.impute
    >>> from sklearn.impute import IterativeImputer

    Args:
        logfilepath (str): Path to the progress log file.

        gridparams (sklearn_genetic.space object or Dict[str, Any]): The parameter distributions or values to use for the grid search.

        clf_kwargs (Dict[str, Any]): A dictionary with the classifier keyword arguments.

        ga_kwargs (Dict[str, Any]): A dictionary with the genetic algorithm arguments.

        prefix (str): Prefix for output files.

        estimator (estimator object, optional): The estimator to use at each step of the round-robin imputation. Defaults to BayesianRidge().

        grid_cv (int, optional): The number of cross-validation folds to use with the grid search. CV folds will be stratified, and it will attempt to balance the genotypes in each fold. IMPORTANT: Sites with fewer than ``2 * grid_cv`` of each genotypes will be removed prior to doing the random subsetting because otherwise the imputation would fail. At least two of each gentoype are required to be present in each fold. Defaults to 5.

        grid_n_jobs (int, optional): [The number of processors to use with the grid search. Set ``grid_n_jobs`` to -1 to use all available processors]. Defaults to 1.

        grid_iter (int, optional): The number of iterations (for RandomizedSearchCV) or generations (for the genetic algorithm) to run with the grid search. For the genetic algorithm, an early stopping callback method is implemented that will stop if the accuracy does not improve for more than 5 consecutive generations. Defaults to 10.

        clf_type (str, optional): Whether to run ``"classifier"`` or ``"regression"`` based imputation. Defaults to "classifier".

        ga (bool, optional): Whether or not to use the genetic algorithm. If True, does genetic algorithm, otherwise does RandomizedSearchCV. Defaults to False.

        disable_progressbar (bool, optional): Whether or not to disable the tqdm progress bar. If True, disables the progress bar. If False, tqdm is used for the progress bar. This can be useful if you are running the imputation on an HPC cluster or are saving the standard output to a file. If True, progress updates will be printed to the screen every ``progress_update_percent`` iterations. Defaults to False.

        progress_update_percent (int, optional) : How often to display progress updates (as a percentage) if ``disable_progressbar`` is True. If ``progress_update_percent=10``\, then it displays progress updates every 10%. Defaults to 10.

        pops (List[Union[int, str]]): List of population IDs of shape (n_samples,).

        scoring_metric (str, optional): Scoring metric to use for the grid search. Defaults to "accuracy".

        early_stop_gen (int, optional): Number of consecutive generations lacking improvement for which to implement the early stopping callback. Defaults to 5.

        missing_values (int or np.nan, optional): The placeholder for the missing values. All occurrences of ``missing_values`` will be imputed. For pandas dataframes with	nullable integer dtypes with missing values, ``missing_values`` should be set to ``np.nan``\, since ``pd.NA`` will be converted to ``np.nan``\. Defaults to np.nan.

        Sample_posterior (bool, optional): CURRENTLY NOT SUPPORTED. Whether to sample from the (Gaussian) predictive posterior of the fitted estimator for each imputation. Estimator must support ``return_std`` in its ``predict`` method if set to ``True``\. Set to ``True`` if using ``IterativeImputer`` for multiple imputations. Defaults to False.

        max_iter (int, optional): Maximum number of imputation rounds to perform before returning the imputations computed during the final round. A round is a single	imputation of each feature with missing values. The stopping criterion is met once ``max(abs(X_t - X_{t-1}))/max(abs(X[known_vals])) < tol``\,	where ``X_t`` is ``X`` at iteration `t`. Note that early stopping is only applied if ``sample_posterior=False``\. Defaults to 10.

        tol (float, optional): Tolerance of the stopping condition. Defaults to 1e-3.

        n_nearest_features (int, optional): Number of other features to use to estimate the missing values of each feature column. Nearness between features is measured using the absolute correlation coefficient between each feature pair (after initial imputation). To ensure coverage of features throughout the imputation process, the neighbor features are not necessarily nearest,	but are drawn with probability proportional to correlation for each	imputed target feature. Can provide significant speed-up when the number of features is huge. If ``None``\, all features will be used. Defaults to None.

        initial_strategy (str, optional): Which strategy to use to initialize the missing values. Same as the ``strategy`` parameter in :class:`~sklearn.impute.SimpleImputer`	Valid values: "most_frequent", "populations", "mf", or "phylogeny". Defaults to "populations".

        imputation_order (str, optional): The order in which the features will be imputed. Possible values: "ascending" (From features with fewest missing values to most), "descending" (From features with most missing values to fewest, "roman" (Left to right), "arabic" (Right to left),  random" (A random order for each round). Defaults to 'ascending'.

        skip_complete (bool, optional): If True then features with missing values during ``transform`` that did not have any missing values during ``fit`` will be imputed with the initial imputation method only. Set to ``True`` if you have	many features with no missing values at both ``fit`` and ``transform`` time to save compute. Defaults to False.

        min_value (float or array-like of shape (n_features,), optional): Minimum possible imputed value. Broadcast to shape (n_features,) if scalar. If array-like, expects shape (n_features,), one min value for each feature. The default is `-np.inf`...versionchanged:: 0.23 (Added support for array-like). Defaults to -np.inf.

        max_value (float or array-like of shape (n_features,), optional): Maximum possible imputed value. Broadcast to shape (n_features,) if scalar. If array-like, expects shape (n_features,), one max value for each feature..versionchanged:: 0.23 (Added support for array-like). Defaults to np.inf.

        verbose (int, optional): Verbosity flag, controls the debug messages that are issued as functions are evaluated. The higher, the more verbose. Can be 0, 1, or 2. Defaults to 0.

        random_state (int or RandomState instance, optional): The seed of the pseudo random number generator to use. Randomizes selection of estimator features if n_nearest_features is not None, the ``imputation_order`` if ``random``\, and the sampling from posterior if ``sample_posterior`` is True. Use an integer for determinism. Defaults to None.

        add_indicator (bool, optional): If True, a :class:`MissingIndicator` transform will stack onto output of the imputer's transform. This allows a predictive estimator to account for missingness despite imputation. If a feature has no missing values at fit/train time, the feature won't appear on the missing indicator even if there are missing values at transform/test time. Defaults to False.

        genotype_data (GenotypeData object, optional): GenotypeData object containing dictionary with keys=sampleIds and values=list of genotypes for the corresponding key. If using ``initial_strategy="phylogeny``\, then this object also needs contain the treefile and qmatrix objects. Defaults to None.

        str_encodings (Dict[str, int], optional): Integer encodings used in STRUCTURE-formatted file. Should be a dictionary with keys=nucleotides and values=integer encodings. The missing data encoding should also be included. Argument is ignored if using a PHYLIP-formatted file. Defaults to {"A": 1, "C": 2, "G": 3, "T": 4, "N": -9}

    Attributes:
        initial_imputer_: (:class:`~sklearn.impute.SimpleImputer`):  Imputer used to initialize the missing values.

        imputation_sequence_ (List[Tuple[numpy.ndarray]]): Each tuple has ``(feat_idx, neighbor_feat_idx, estimator)``\, where ``feat_idx`` is the current feature to be imputed, ``neighbor_feat_idx`` is the array of other features used to impute the current feature, and ``estimator`` is the trained estimator used for the imputation. Length is ``self.n_features_with_missing_ *	self.n_iter_``\.

        n_iter_ (int): Number of iteration rounds that occurred. Will be less than ``self.max_iter`` if early stopping criterion was reached.

        n_features_with_missing_ (int): Number of features with missing values.

        indicator_ (:class:`~sklearn.impute.MissingIndicator`): Indicator used to add binary indicators for missing values ``None`` if add_indicator is False.

        random_state_ (RandomState instance): RandomState instance that is generated either from a seed, the random number generator or by ``np.random``\.

        genotype_data (GenotypeData object): GenotypeData object.

        str_encodings (Dict[str, int]): Dictionary with integer encodings for converting from STRUCTURE-formatted file to IUPAC nucleotides.

    See Also:
        SimpleImputer : Univariate imputation of missing values.

    Examples:
        >>> import numpy as np
        >>> from sklearn.experimental import enable_iterative_imputer
        >>> from sklearn.impute import IterativeImputer
        >>> imp_mean = IterativeImputer(random_state=0)
        >>> imp_mean.fit([[7, 2, 3], [4, np.nan, 6], [10, 5, 9]])
        >>> X = [[np.nan, 2, 3], [4, np.nan, 6], [10, np.nan, 9]]
        >>> imp_mean.transform(X)
        array([[ 6.9584...,  2.       ,  3.        ],
                        [ 4.       ,  2.6000...,  6.        ],
                        [10.       ,  4.9999...,  9.        ]])

    Notes:
        To support imputation in inductive mode we store each feature's estimator during the ``fit`` phase, and predict without refitting (in order) during	the ``transform`` phase. Features which contain all missing values at ``fit`` are discarded upon ``transform``\.

    References:
        .. [van Buuren & Groothuis-Oudshoorn, 2011] Stef van Buuren, Karin Groothuis-Oudshoorn (2011). mice: Multivariate Imputation by Chained Equations in R. Journal of Statistical Software 45: 1-67.

        .. [Buck, 1960] S. F. Buck. (1960). A Method of Estimation of Missing Values in	Multivariate Data Suitable for use with an Electronic Computer. Journal of the Royal Statistical Society 22(2): 302-306.
    """

    def __init__(
        self,
        logfilepath: str,
        clf_kwargs: Dict[str, Any],
        ga_kwargs: Dict[str, Any],
        *,
        estimator: Callable = None,
        gridparams: Dict[str, Any] = None,
        prefix: str = "output",
        grid_cv: int = 5,
        grid_n_jobs: int = 1,
        grid_iter: int = 10,
        clf_type: str = "classifier",
        gridsearch_method: str = "gridsearch",
        disable_progressbar: bool = False,
        progress_update_percent: Optional[int] = None,
        pops: Optional[List[Union[str, int]]] = None,
        scoring_metric: str = "accuracy",
        early_stop_gen: int = 5,
        missing_values: Union[int, float] = np.nan,
        sample_posterior: bool = False,
        max_iter: int = 10,
        tol: float = 1e-3,
        n_nearest_features: Optional[int] = None,
        initial_strategy: str = "populations",
        imputation_order: str = "ascending",
        skip_complete: bool = False,
        min_value: Union[float, int, float] = -np.inf,
        max_value: Union[float, int, float] = np.inf,
        verbose: int = 0,
        random_state: Optional[int] = None,
        add_indicator: bool = False,
        genotype_data: Optional[Any] = None,
        str_encodings: Optional[Dict[str, int]] = None,
    ) -> None:
        super().__init__(
            estimator=estimator,
            missing_values=missing_values,
            sample_posterior=sample_posterior,
            max_iter=max_iter,
            tol=tol,
            n_nearest_features=n_nearest_features,
            initial_strategy=initial_strategy,
            imputation_order=imputation_order,
            skip_complete=skip_complete,
            min_value=min_value,
            max_value=max_value,
            verbose=verbose,
            random_state=random_state,
            add_indicator=add_indicator,
        )

        self.logfilepath = logfilepath
        self.gridparams = gridparams
        self.clf_kwargs = clf_kwargs
        self.ga_kwargs = ga_kwargs
        self.prefix = prefix
        self.estimator = estimator
        self.sample_posterior = sample_posterior
        self.max_iter = max_iter
        self.tol = tol
        self.n_nearest_features = n_nearest_features
        self.initial_strategy = initial_strategy
        self.imputation_order = imputation_order
        self.skip_complete = skip_complete
        self.min_value = min_value
        self.max_value = max_value
        self.verbose = verbose
        self.random_state = random_state
        self.genotype_data = genotype_data
        self.str_encodings = str_encodings
        self.grid_cv = grid_cv
        self.grid_n_jobs = grid_n_jobs
        self.grid_iter = grid_iter
        self.clf_type = clf_type
        self.gridsearch_method = gridsearch_method
        self.disable_progressbar = disable_progressbar
        self.progress_update_percent = progress_update_percent
        self.pops = pops
        self.scoring_metric = scoring_metric
        self.early_stop_gen = early_stop_gen
        self.missing_values = missing_values

    def _mode_1d(self, array_1d):
        """Get the mode of a 1D array.

        Args:
            array_1d (np.ndarray): 1D array to calculate the mode for.

        Returns:
            int: Most common value of the 1D array.
        """
        # Get unique elements and their corresponding counts
        vals, counts = np.unique(array_1d, return_counts=True)

        # Get the index of the most frequent element
        mode_idx = np.argmax(counts)

        # Return the mode
        return vals[mode_idx]

    # Define a function to remove columns with a unique count of 1 or 0
    def remove_single_unique_val_cols(self, array_2d, return_idx=False):
        # Apply np.unique to each column and get the counts of unique values (excluding np.nan)
        unique_counts = np.apply_along_axis(
            lambda x: len(np.unique(x[~np.isnan(x)])), axis=0, arr=array_2d
        )

        # Get the column indices where the count of unique values is greater than 1
        cols_to_keep = np.where(unique_counts > 1)[0]

        # Index into array_2d to keep only these columns
        array_2d_filtered = array_2d[:, cols_to_keep]

        if return_idx:
            return cols_to_keep
        else:
            return array_2d_filtered

    def _initial_imputation(
        self, X: np.ndarray, cols_to_keep: np.ndarray, in_fit: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Perform initial imputation for input X.

        Initially imputes training data using a simple imputation method.

        Args:
            X (numpy.ndarray, shape (n_samples, n_features)): Input data, where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

            cols_to_keep (numpy.ndarray, shape (n_features,)): Column indices to keep. Only used if ``initial_strategy="phylogeny"``\.

            in_fit (bool, optional): Whether function is called in fit. Defaults to False.

        Returns:
            Xt (numpy.ndarray, shape (n_samples, n_features): Input data, where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

            X_filled (numpy.ndarray, shape (n_samples, n_features)): [Input data with the most recent imputations.

            mask_missing_values (numpy.ndarray, shape (n_samples, n_features)): Input data's missing indicator matrix, where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

            X_missing_mask (numpy.ndarray, shape (n_samples, n_features)): Input data's mask matrix indicating missing datapoints, where
                ``n_samples`` is the number of samples and ``n_features`` is the
                number of features.
        """
        if is_scalar_nan(self.missing_values):
            force_all_finite = "allow-nan"
        else:
            force_all_finite = True

        X = self._validate_data(
            X,
            dtype=FLOAT_DTYPES,
            order="F",
            reset=in_fit,
            force_all_finite=force_all_finite,
        )

        X[X < 0] = np.nan

        _check_inputs_dtype(X, self.missing_values)

        X_missing_mask = _get_mask(X, self.missing_values)
        mask_missing_values = X_missing_mask.copy()

        if self.initial_strategy == "populations":
            self.initial_imputer_ = simple_imputers.ImputeAlleleFreq(
                self.genotype_data,
                pops=self.pops,
                by_populations=True,
                missing=-9,
                write_output=False,
                output_format="array",
                verbose=False,
                iterative_mode=True,
                validation_mode=True,
            )

            X_filled = self.initial_imputer_.imputed

        elif self.initial_strategy == "phylogeny":
            if (
                self.genotype_data.qmatrix is None
                and self.genotype_data.qmatrix_iqtree is None
            ) or self.genotype_data.guidetree is None:
                raise AttributeError(
                    "GenotypeData object was not initialized with "
                    "qmatrix/ qmatrix_iqtree or guidetree arguments, "
                    "but initial_strategy == 'phylogeny'"
                )

            else:
                self.initial_imputer_ = simple_imputers.ImputePhylo(
                    genotype_data=self.genotype_data,
                    str_encodings=self.str_encodings,
                    write_output=False,
                    disable_progressbar=True,
                    column_subset=cols_to_keep,
                    validation_mode=True,
                )

                X_filled = self.initial_imputer_.imputed
                valid_sites = self.initial_imputer_.valid_sites

        elif self.initial_strategy == "mf":
            self.initial_imputer_ = simple_imputers.ImputeMF(
                self.genotype_data,
                missing=-9,
                write_output=False,
                verbose=False,
                output_format="array",
                validation_mode=True,
            )

            X_filled = self.initial_imputer_.imputed

        else:
            if self.initial_imputer_ is None:
                self.initial_imputer_ = SimpleImputer(
                    missing_values=self.missing_values,
                    strategy=self.initial_strategy,
                )
                X_filled = self.initial_imputer_.fit_transform(X)

            else:
                X_filled = self.initial_imputer_.transform(X)

            valid_sites = self.initial_imputer_.statistics_

        Xt = X.copy()

        if self.initial_imputer_ == "phylogeny":
            valid_sites = np.apply_along_axis(self._mode_1d, axis=0, arr=Xt)
            valid_sites = valid_sites[cols_to_keep]
            valid_mask = np.flatnonzero(np.logical_not(np.isnan(valid_sites)))
            Xt = X[:, valid_mask]
            mask_missing_values = mask_missing_values[:, valid_mask]
            X_filled = X_filled[:, valid_mask]

        return Xt, X_filled, mask_missing_values, X_missing_mask

    @ignore_warnings(category=UserWarning)
    def _impute_one_feature(
        self,
        X_filled: np.ndarray,
        mask_missing_values: np.ndarray,
        feat_idx: int,
        neighbor_feat_idx: np.ndarray,
        estimator: Optional[Any] = None,
        fit_mode: bool = True,
    ) -> Tuple[np.ndarray, Optional[Any]]:
        """Impute a single feature from the others provided.

        This function predicts the missing values of one of the features using the current estimates of all the other features. The ``estimator`` must support ``return_std=True`` in its ``predict`` method for this function to work.

        Args:
            X_filled (numpy.ndarray): Input data with the most recent imputations.

            mask_missing_values (numpy.ndarray): Input data's missing indicator matrix.

            feat_idx (int): Index of the feature currently being imputed.

            neighbor_feat_idx (numpy.ndarray): Indices of the features to be used in imputing ``feat_idx``\.

            estimator (sklearn estimator object, optional): The estimator to use at this step of the round-robin imputation If ``sample_posterior`` is True, the estimator must support ``return_std`` in its ``predict`` method. If None, it will be cloned from self._estimator. Defaults to None.

            fit_mode (bool, optional): Whether to fit and predict with the estimator or just predict. Defaults to True.

        Returns:
            X_filled (ndarray): Input data with ``X_filled[missing_row_mask, feat_idx]`` updated.

            estimator (estimator with sklearn API): The fitted estimator used to impute ``X_filled[missing_row_mask, feat_idx]``\.
        """
        if estimator is None and fit_mode is False:
            raise ValueError(
                "If fit_mode is False, then an already-fitted "
                "estimator should be passed in."
            )

        if type(self._estimator).__name__ == "XGBClassifier":
            self.gridparams["objective"] = ["multi:softmax"]
            self.gridparams["num_class"] = [3]

        if estimator is None:
            estimator = clone(self._estimator)

        # Modified code
        cross_val = StratifiedKFold(n_splits=self.grid_cv, shuffle=True)

        # Modified code
        # If regressor
        if self.clf_type == "regressor":
            if self.gridsearch_method == "genetic_algorithm":
                callback = DeltaThreshold(threshold=1e-3, metric="fitness")

        else:
            if self.gridsearch_method == "genetic_algorithm":
                callback = ConsecutiveStopping(
                    generations=self.early_stop_gen, metric="fitness"
                )

        # Do randomized grid search
        if self.gridsearch_method == "randomized_gridsearch":
            search = RandomizedSearchCV(
                estimator,
                param_distributions=self.gridparams,
                n_iter=self.grid_iter,
                scoring=self.scoring_metric,
                n_jobs=self.grid_n_jobs,
                cv=cross_val,
            )

        elif self.gridsearch_method == "gridsearch":
            search = GridSearchCV(
                estimator,
                param_grid=self.gridparams,
                scoring=self.scoring_metric,
                n_jobs=self.grid_n_jobs,
                cv=cross_val,
                error_score="raise",
            )

        # Do genetic algorithm
        elif self.gridsearch_method == "genetic_algorithm":
            with HiddenPrints():
                search = GASearchCV(
                    estimator=estimator,
                    cv=cross_val,
                    scoring=self.scoring_metric,
                    generations=self.grid_iter,
                    param_grid=self.gridparams,
                    n_jobs=self.grid_n_jobs,
                    verbose=False,
                    **self.ga_kwargs,
                )

        else:
            raise ValueError(
                f"Invalid gridsearch_method provided: {self.gridsearch_method}. Supported options are 'gridsearch', 'randomized_gridsearch', and 'genetic_algorithm'"
            )

        missing_row_mask = mask_missing_values[:, feat_idx]

        if fit_mode:
            X_train = _safe_indexing(
                X_filled[:, neighbor_feat_idx], ~missing_row_mask
            )

            y_train = _safe_indexing(X_filled[:, feat_idx], ~missing_row_mask)
            X_train = X_train.astype(int)
            y_train = y_train.astype(int)

            if self.gridsearch_method == "genetic_algorithm":
                if type(self._estimator).__name__ == "XGBClassifier":
                    raise NotImplementedError(
                        "genetic_algorithm is not currently supported with ImputeXGBoost."
                    )
                search.fit(X_train, y_train, callbacks=callback)
            else:
                if type(self._estimator).__name__ == "XGBClassifier":
                    oe = OrdinalEncoder(
                        categories="auto",
                        dtype=int,
                        handle_unknown="use_encoded_value",
                        unknown_value=-1,
                    )
                    X_train = oe.fit_transform(X_train)

                    # Add one dummy sample for each possible class
                    for class_label in range(
                        3
                    ):  # assuming there are 3 classes 0, 1, 2
                        class_count = np.count_nonzero(y_train == class_label)
                        if (
                            class_label not in y_train
                            or class_count < self.grid_cv
                        ):
                            for _ in range(
                                self.grid_cv - class_count
                            ):  # add as many dummy samples as there are folds
                                dummy_sample = np.zeros((1, X_train.shape[1]))
                                X_train = np.vstack([X_train, dummy_sample])
                                y_train = np.append(y_train, class_label)

                    # Fit the LabelEncoder after adding the dummy samples
                    le = LabelEncoder()
                    y_train = le.fit_transform(y_train)
                    sample_weight = compute_sample_weight("balanced", y_train)

                    search.fit(X_train, y_train, sample_weight=sample_weight)
                else:
                    search.fit(X_train, y_train)

        # if no missing values, don't predict
        if np.sum(missing_row_mask) == 0:
            return X_filled, None

        X_test = _safe_indexing(
            X_filled[:, neighbor_feat_idx], missing_row_mask
        )

        X_test = X_test.astype(int)

        if type(self._estimator).__name__ == "XGBClassifier":
            X_test = oe.transform(X_test)

        # Currently un-tested with grid search
        if self.sample_posterior:
            raise NotImplementedError(
                "sample_posterior is not implemented in PG-SUI"
            )

        else:
            imputed_values = search.predict(X_test)

            if type(self._estimator).__name__ == "XGBClassifier":
                imputed_values = le.inverse_transform(imputed_values)

            imputed_values = np.clip(
                imputed_values,
                self._min_value[feat_idx],
                self._max_value[feat_idx],
            )

        # update the feature
        X_filled[missing_row_mask, feat_idx] = imputed_values

        del X_train
        del y_train
        del X_test
        del imputed_values
        gc.collect()

        return X_filled, search

    @ignore_warnings(category=UserWarning)
    def fit_transform(
        self,
        X: np.ndarray,
        valid_cols: Optional[np.ndarray] = None,
        y: None = None,
    ) -> Tuple[np.ndarray, Optional[List[Any]], Optional[List[Any]]]:
        """Fits the imputer on X and return the transformed X.

        The basic functionality is to get the nearest neightbors for each feature (column) and loop through all the features and their correlated neighbors to predict missing values.

        Functionality has been added to perform grid searches using two methods (genetic algorithm and RandomSearchCV). It also makes several useful plots if using the genetic algorithm, and a tqdm progress bar and status updates have been added.

        Args:
            X (array-like, shape (n_samples, n_features)): Input data, where "n_samples" is the number of samples and "n_features" is the number of features.

            valid_cols (numpy.ndarray, optional): Array with column indices to keep. Defaults to None.

            y (None, optional): Ignored. Here for compatibility with other sklearn classes. Defaults to None.

        Returns:
            array-like, shape (n_samples, n_features)): The imputed input data.
            List[Union[str, int, float]]: List of parameter settings found.
            List[Union[int, float]]: List of scores.
        """
        self.random_state_ = getattr(
            self, "random_state_", check_random_state(self.random_state)
        )

        if self.max_iter < 0:
            raise ValueError(
                f"'max_iter' should be a positive integer. Got {self.max_iter} instead."
            )

        if self.tol < 0:
            raise ValueError(
                f"'tol' should be a non-negative float. Got {self.tol} instead"
            )

        self._estimator = clone(self.estimator)

        self.initial_imputer_ = None

        X, Xt, mask_missing_values, complete_mask = self._initial_imputation(
            X.copy(), valid_cols, in_fit=True
        )

        super(IterativeImputer, self)._fit_indicator(complete_mask)
        X_indicator = super(IterativeImputer, self)._transform_indicator(
            complete_mask
        )

        if self.max_iter == 0 or np.all(mask_missing_values):
            self.n_iter_ = 0
            return (
                super(IterativeImputer, self)._concatenate_indicator(
                    Xt, X_indicator
                ),
                None,
                None,
            )

        # Edge case: a single feature. We return the initial ...
        if Xt.shape[1] == 1:
            self.n_iter_ = 0
            return (
                super(IterativeImputer, self)._concatenate_indicator(
                    Xt, X_indicator
                ),
                None,
                None,
            )

        self._min_value = self._validate_limit(
            self.min_value, "min", X.shape[1]
        )
        self._max_value = self._validate_limit(
            self.max_value, "max", X.shape[1]
        )

        if not np.all(np.greater(self._max_value, self._min_value)):
            raise ValueError(
                "One (or more) features have min_value >= max_value."
            )

        # order in which to impute
        # note this is probably too slow for large feature data (d > 100000)
        # and a better way would be good.
        # see: https://goo.gl/KyCNwj and subsequent comments
        ordered_idx = self._get_ordered_idx(mask_missing_values)
        total_features = len(ordered_idx)

        self.n_features_with_missing_ = len(ordered_idx)

        abs_corr_mat = self._get_abs_corr_mat(Xt)

        _, n_features = Xt.shape

        if self.verbose > 0:
            print(
                f"[IterativeImputer] Completing matrix with shape ({X.shape},)"
            )
        start_t = time()

        if not self.sample_posterior:
            Xt_previous = Xt.copy()
            normalized_tol = self.tol * np.max(np.abs(X[~mask_missing_values]))

        params_list = list()
        score_list = list()
        iter_list = list()

        if self.gridsearch_method == "genetic_algorithm":
            sns.set_style("white")

        total_iter = self.max_iter

        #######################################
        ### Iterations
        #######################################
        for self.n_iter_ in progressbar(
            range(1, total_iter + 1),
            desc="Iteration: ",
            disable=self.disable_progressbar,
        ):
            if self.gridsearch_method == "genetic_algorithm":
                iter_list.append(self.n_iter_)

                pp_oneline = PdfPages(
                    f".score_traces_separate_{self.n_iter_}.pdf"
                )

                pp_lines = PdfPages(
                    f".score_traces_combined_{self.n_iter_}.pdf"
                )

                pp_space = PdfPages(f".search_space_{self.n_iter_}.pdf")

            if self.imputation_order == "random":
                ordered_idx = self._get_ordered_idx(mask_missing_values)

            # Reset lists for current iteration
            params_list.clear()
            score_list.clear()
            searches = list()

            if self.disable_progressbar:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(
                            f"Iteration Progress: {self.n_iter_}/{self.max_iter} ({int((self.n_iter_ / total_iter) * 100)}%)"
                        )

                if self.progress_update_percent is not None:
                    print_perc_interval = self.progress_update_percent

            ########################################
            ### Features
            ########################################
            for i, feat_idx in enumerate(
                progressbar(
                    ordered_idx,
                    desc="Feature: ",
                    leave=False,
                    position=1,
                    disable=self.disable_progressbar,
                ),
                start=1,
            ):
                neighbor_feat_idx = self._get_neighbor_feat_idx(
                    n_features, feat_idx, abs_corr_mat
                )

                Xt, search = self._impute_one_feature(
                    Xt,
                    mask_missing_values,
                    feat_idx,
                    neighbor_feat_idx,
                    estimator=None,
                    fit_mode=True,
                )

                searches.append(search)

                # NOTE: The below source code has been commented out to save
                # RAM. estimator_triplet object contains numerous fit estimators
                # that demand a lot of resources. It is primarily used for the
                # transform function, which is not needed in this application.

                # estimator_triplet = _ImputerTripletGrid(feat_idx,
                # 									neighbor_feat_idx,
                # 									estimator)

                # self.imputation_sequence_.append(estimator_triplet)

                if search is not None:
                    # There was missing data in the feature
                    params_list.append(search.best_params_)
                    score_list.append(search.best_score_)

                    if self.gridsearch_method == "genetic_algorithm":
                        plt.cla()
                        plt.clf()
                        plt.close()

                        plot_fitness_evolution(search)
                        pp_oneline.savefig(bbox_inches="tight")
                        plt.cla()
                        plt.clf()
                        plt.close()

                        Plotting.plot_search_space(search)
                        pp_space.savefig(bbox_inches="tight")
                        plt.cla()
                        plt.clf()
                        plt.close()

                else:
                    # Search is None
                    # Thus, there was no missing data in the given feature
                    tmp_dict = dict()
                    for k in self.gridparams.keys():
                        tmp_dict[k] = -9
                    params_list.append(tmp_dict)

                    score_list.append(-9)

                # Only print feature updates at each progress_update_percent
                # interval
                if (
                    self.progress_update_percent is not None
                    and self.disable_progressbar
                ):
                    current_perc = math.ceil((i / total_features) * 100)

                    if current_perc >= print_perc_interval:
                        with open(self.logfilepath, "a") as fout:
                            # Redirect progress to file
                            with redirect_stdout(fout):
                                print(
                                    f"Feature Progress (Iteration "
                                    f"{self.n_iter_}/{self.max_iter}): "
                                    f"{i}/{total_features} ({current_perc}"
                                    f"%)"
                                )

                            if i == len(ordered_idx):
                                with redirect_stdout(fout):
                                    print("")

                        while print_perc_interval <= current_perc:
                            print_perc_interval += self.progress_update_percent

            if self.verbose > 1:
                print(
                    f"[IterativeImputer] Ending imputation round "
                    f"{self.n_iter_}/{self.max_iter}, "
                    f"elapsed time {(time() - start_t):0.2f}"
                )

            if not self.sample_posterior:
                inf_norm = np.linalg.norm(
                    Xt - Xt_previous, ord=np.inf, axis=None
                )
                if self.verbose > 0:
                    print(
                        f"[IterativeImputer] Change: {inf_norm}, "
                        f"scaled tolerance: {normalized_tol} "
                    )

                if inf_norm < normalized_tol:
                    if self.disable_progressbar:
                        # Early stopping criteria has been reached
                        with open(self.logfilepath, "a") as fout:
                            # Redirect to progress logfile
                            with redirect_stdout(fout):
                                print(
                                    "[IterativeImputer] Early stopping "
                                    "criterion reached."
                                )
                    else:
                        print(
                            "[IterativeImputer] Early stopping criterion "
                            "reached."
                        )

                    if self.gridsearch_method == "genetic_algorithm":
                        pp_oneline.close()
                        pp_space.close()

                        plt.cla()
                        plt.clf()
                        plt.close()
                        for iter_search in searches:
                            if iter_search is not None:
                                plot_fitness_evolution(iter_search)

                        pp_lines.savefig(bbox_inches="tight")

                        plt.cla()
                        plt.clf()
                        plt.close()
                        pp_lines.close()

                    break
                Xt_previous = Xt.copy()

            if self.gridsearch_method == "genetic_algorithm":
                pp_oneline.close()
                pp_space.close()

                plt.cla()
                plt.clf()
                plt.close()
                for iter_search in searches:
                    if iter_search is not None:
                        plot_fitness_evolution(iter_search)

                pp_lines.savefig(bbox_inches="tight")

                plt.cla()
                plt.clf()
                plt.close()
                pp_lines.close()

        else:
            if not self.sample_posterior:
                warnings.warn(
                    "[IterativeImputer] Early stopping criterion not"
                    " reached.",
                    ConvergenceWarning,
                )

        Xt[~mask_missing_values] = X[~mask_missing_values]
        Xt = Xt.astype(int)

        if self.gridsearch_method == "genetic_algorithm":
            # Remove all files except last iteration
            final_iter = iter_list.pop()

            [os.remove(f".score_traces_separate_{x}.pdf") for x in iter_list]
            [os.remove(f".score_traces_combined_{x}.pdf") for x in iter_list]
            [os.remove(f".search_space_{x}.pdf") for x in iter_list]

            shutil.move(
                f".score_traces_separate_{final_iter}.pdf",
                f"{self.prefix}_score_traces_separate.pdf",
            )

            shutil.move(
                f".score_traces_combined_{final_iter}.pdf",
                f"{self.prefix}_score_traces_combined.pdf",
            )

            shutil.move(
                f".search_space_{final_iter}.pdf",
                f"{self.prefix}_search_space.pdf",
            )

        return (
            super(IterativeImputer, self)._concatenate_indicator(
                Xt, X_indicator
            ),
            params_list,
            score_list,
        )



import logging
import math
import os
import sys
import random
import warnings

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from sklearn.utils.class_weight import (
    compute_class_weight,
)

from sklearn.metrics import f1_score

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

# noinspection PyPackageRequirements
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns of GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated


class DisabledCV:
    def __init__(self):
        self.n_splits = 1

    def split(self, X, y, groups=None):
        yield (np.arange(len(X)), np.arange(len(y)))

    def get_n_splits(self, X, y, groups=None):
        return self.n_splits


# For VAE.
# Necessary to initialize outside of class for use with tf.function decorator.
cce = tf.keras.losses.CategoricalCrossentropy(from_logits=True)
cca = tf.keras.metrics.CategoricalAccuracy()
ba = tf.keras.metrics.BinaryAccuracy()
bce = tf.keras.losses.BinaryCrossentropy()


class NeuralNetworkMethods:
    """Methods common to all neural network imputer classes and loss functions"""

    def __init__(self):
        self.data = None

    @staticmethod
    def encode_multilab(X, num_classes=4):
        """Encode 0-9 integer data in one-hot format.
        Args:
            X (numpy.ndarray): Input array with 012-encoded data and -9 as the missing data value.

            num_classes (int, optional): Number of multi-label classes to use. Mostly for compatibility with encode_multiclass. Defaults to 4.
        Returns:
            pandas.DataFrame: One-hot encoded data, ignoring missing values (np.nan). multi-label categories will be encoded as 0.5. Otherwise, it will be 1.0.
        """
        # return np.where(X >= 0.5, 1.0, 0.0)
        try:
            Xt = np.zeros(shape=(X.shape[0], X.shape[1], 4))
        except IndexError:
            Xt = np.zeros(shape=(X.shape[0],))

        mappings = {
            0: [1.0, 0.0, 0.0, 0.0],
            1: [0.0, 1.0, 0.0, 0.0],
            2: [0.0, 0.0, 1.0, 0.0],
            3: [0.0, 0.0, 0.0, 1.0],
            4: [1.0, 1.0, 0.0, 0.0],
            5: [1.0, 0.0, 1.0, 0.0],
            6: [1.0, 0.0, 0.0, 1.0],
            7: [0.0, 1.0, 1.0, 0.0],
            8: [0.0, 1.0, 0.0, 1.0],
            9: [0.0, 0.0, 1.0, 1.0],
            -9: [np.nan, np.nan, np.nan, np.nan],
        }
        try:
            for row in np.arange(X.shape[0]):
                Xt[row] = [mappings[enc] for enc in X[row]]
        except TypeError:
            Xt = [mappings[enc] for enc in X]

        if not isinstance(Xt, np.ndarray):
            Xt = np.array(Xt)
        return Xt

    @staticmethod
    def encode_multiclass(X, num_classes=10, missing_value=-9):
        """Encode 0-9 integer data in multi-class one-hot format.

        Missing values get encoded as ``[np.nan] * num_classes``
        Args:
            X (numpy.ndarray): Input array with 012-encoded data and ``missing_value`` as the missing data value.

            num_classes (int, optional): Number of classes to use. Defaults to 10.

            missing_value (int, optional): Missing data value to replace with ``[np.nan] * num_classes``\. Defaults to -9.
        Returns:
            pandas.DataFrame: Multi-class one-hot encoded data, ignoring missing values (np.nan).
        """
        int_cats, ohe_arr = np.arange(num_classes), np.eye(num_classes)
        mappings = dict(zip(int_cats, ohe_arr))
        mappings[missing_value] = np.array([np.nan] * num_classes)

        try:
            Xt = np.zeros(shape=(X.shape[0], X.shape[1], num_classes))
        except IndexError:
            Xt = np.zeros(shape=(X.shape[0],))

        try:
            for row in np.arange(X.shape[0]):
                Xt[row] = [mappings[enc] for enc in X[row]]
        except TypeError:
            Xt = [mappings[enc] for enc in X]

        if not isinstance(Xt, np.ndarray):
            Xt = np.array(Xt)

        return Xt

    @classmethod
    def decode_masked(
        cls,
        y_true_bin,
        y_pred_proba,
        is_multiclass=True,
        return_proba=False,
        return_multilab=False,
        return_int=True,
        predict_still_missing=True,
        threshold_increment=0.01,
        multilabel_averaging="macro",
        missing_mask=None,
    ):
        """Evaluate model predictions by decoding from one-hot encoding to integer-encoded format.

        Gets the index of the highest predicted value to obtain the integer encodings or integer encodings.

        Calucalates highest predicted value for each row vector and each class, setting the most likely class to 1.0.

        Args:
            y_true_bin (numpy.ndarray): True multilabel target values of shape (n_samples * n_features, num_classes). Array should be flattened and masked.

            y_pred_proba (numpy.ndarray): Multilabel model predictions of shape (n_samples * n_features, num_classes). Array should be flattened and masked.

            is_multiclass (bool, optional): True if using multiclass data with softmax activation. False if using multilabel data with sigmoid activation. Defaults to True.

            threshold (float, optional): If using multilabel, then set the threshold for determining 1 or 0 predictions. Defaults to 0.5.

            return_proba (bool, optional): If True, returns probabilities for unresolved values where all multilabel probabilities were below the threshold. Defaults to False.

            return_multilab (bool, optional): If True, returns the multilabel encodings instead of integer encodings (if doing multilabel classification). Defaults to False.

            return_int (bool, optional): If True, returns the integer encodings instead of onehot encodings (if doing multiclass classification). Defaults to False.

            predict_still_missing (bool, optional): If True, values that are still missing after decoding are decoded using the maximum probability (i.e., with np.argmax). If False, then it is possible that some missing data might still remain after decoding if none of the multilabel probabilities are above the threshold. Defaults to True.

            threshold_increment (float, optional): How much to increment threshold when searching for optimal threshold. Should be > 0 and < 1. Defaults to 0.05.

            multilabel_averaging (str): Method to use for averaging F1 score among multilabel classes. Supported options are: {"macro", "micro", "weighted", "samples"}. Defaults to "macro".

            missing_mask (numpy.ndarray, optional): Missing mask with missing values encoded as 1's and nonmissing as 0. Only used if not None. Defaults to None.

        Returns:
            numpy.ndarray: Imputed integer-encoded values.

            numpy.ndarray (optional): Probabilities for each call, with those above the threshold set to 1.0 and those below the threshold between 0 and 1.
        """

        if return_int and return_multilab:
            raise ValueError(
                "return_int and return_multilab cannot both be True."
            )

        y_unresolved_certainty = None
        if is_multiclass or y_true_bin.shape[-1] == 10:
            # Softmax predictions.
            # If reduce_dim is True, will return integer encodings.
            # Otherwise, returns one-hot encodings.
            y_pred = cls.decode_multiclass(y_pred_proba, reduce_dim=return_int)
        else:
            # Onehot encode if not already one-hot encoded.
            if y_true_bin.shape[-1] != 4:
                if y_true_bin.shape[-1] != 10:
                    y_true_bin = cls.encode_multilab(y_true_bin)
                else:
                    y_true_bin = cls.encode_multiclass(y_true_bin)

            pred_multilab = cls.zero_extra_categories(y_pred_proba)

            # Binary multilabel predictions.
            threshold = cls.get_optimal_threshold(
                y_true_bin,
                pred_multilab,
                increment=threshold_increment,
                average_method=multilabel_averaging,
            )

            # Call 0s and 1s based on threshold.
            pred_multilab = np.where(pred_multilab >= threshold, 1.0, 0.0)

            pred_multilab_decoded = cls.decode_binary_multilab(pred_multilab)

            if predict_still_missing:
                # Check if there are still any missing values.
                still_missing = np.all(pred_multilab == 0, axis=-1)

                if return_multilab:
                    still_missing_bin = np.all(
                        pred_multilab == 0, axis=-1, keepdims=True
                    )

                # Do multiclass prediction with argmax then get the probabilities
                # if any unresolved values.
                if np.any(still_missing):
                    # Get the argmax with the highest probability if
                    # all classes are below threshold.
                    y_multi = cls.decode_multiclass(y_pred_proba)
                    y_multi_bin = cls.decode_multiclass(
                        y_pred_proba, reduce_dim=False
                    )

                    try:
                        y_pred = np.where(
                            still_missing, y_multi, pred_multilab_decoded
                        )
                    except ValueError:
                        y_pred = np.where(
                            still_missing,
                            y_multi,
                            np.reshape(
                                pred_multilab_decoded, still_missing.shape
                            ),
                        )

                    if return_multilab:
                        y_pred_bin = np.where(
                            still_missing_bin, y_multi_bin, pred_multilab
                        )

                        y_pred = y_pred_bin

                    if return_proba:
                        # Get max value as base call.
                        y_pred_proba_max = y_pred_proba.max(axis=-1)

                        # Get probability of max value that was < threshold.
                        y_unresolved_certainty = np.where(
                            still_missing, y_pred_proba_max, 1.0
                        )

                else:
                    if return_multilab:
                        y_pred = pred_multilab
                    else:
                        y_pred = pred_multilab_decoded
            else:
                if return_multilab:
                    y_pred = pred_multilab
                else:
                    y_pred = pred_multilab_decoded

        y_pred = y_pred.astype(int)

        if return_proba:
            return y_pred, y_unresolved_certainty
        else:
            return y_pred

    @classmethod
    def get_optimal_threshold(
        cls,
        y_true_bin,
        y_pred_proba,
        increment=0.01,
        average_method="macro",
    ):
        """Increment to find the optimal decoding threshold.

        Args:
            y_true_bin (numpy.ndarray): True multilabel values of shape (n_samples * n_features, num_classes).

            y_pred_proba (numpy.ndarray): Multilabel prediction probabilities of shape (n_features * n_samples, num_classes).

            increment (float, optional): How much to increment when searching for optimal threshold. Should be > 0 and < 1. Defaults to 0.1.

            average_method (str, optional): Method to use for averaging the F1 score across multilabel classes. Possible options include {"macro", "micro", "weighted", "samples"}. Defaults to "macro".

        Returns:
            float: Optimal decoding threshold.
        """
        y_true = y_true_bin.copy()
        y_pred = y_pred_proba.copy()

        thresholds = np.arange(increment, 1, increment)

        nonmissing_mask = np.where(y_true_bin != -1)
        num_classes = y_true_bin.shape[-1]

        # This is only supposed to get applied during the final transform,
        # when the original missing data is replaced with predictions.
        # If this isn't done here, it ends up having -1 values in it,
        # which causes the f1_score function to throw an error.

        try:
            y_true = y_true[nonmissing_mask]
            y_pred = y_pred[nonmissing_mask]
        except IndexError:
            pass

        # Call 0s and 1s based on threshold.

        scores = list()
        for t in thresholds:
            pred_multilab = np.where(y_pred >= t, 1.0, 0.0)
            pred_multilab_decoded = cls.decode_binary_multilab(pred_multilab)
            true_multilab_decoded = cls.decode_binary_multilab(y_true)

            # Had to cast them as integers to get rid of a type error during the
            # final transform() function.

            scores.append(
                f1_score(
                    true_multilab_decoded,
                    pred_multilab_decoded,
                    average="weighted",
                )
            )

        return thresholds[np.argmax(scores)]

    @classmethod
    def flatten_bin_encodings(cls, y):
        """Flatten first two dimensions of binary encodings to (num_samples * num_features, num_classes).

        Args:
            y (numpy.ndarray): Numpy array with 3-dimensional shape of (n_samples, num_features, num_classes).

        Returns:
            numpy.ndarray: Array of shape (n_samples * num_features, num_classes).

        Raises:
            ValueError: Input shape must be 3-dimensional.
        """
        if len(y.shape) != 3:
            raise ValueError("Input array must be 3-dimensional")

        return y.reshape(y.shape[0] * y.shape[1], y.shape[2])

    @staticmethod
    def zero_extra_categories(y_pred_proba, threshold=0.5):
        """Check if any prediction probabilities have >2 values above threshold.

        If >2, then it sets the two with the lowest probabilities to 0.0.

        Args:
            y_pred_proba (numpy.ndarray): Prediction probabilities (sigmoid activation) of shape (n_samples, n_features, num_classes) or (n_samples * n_features, num_classes).

            pred_multilab (numpy.ndarray): Multi-label decodings. Inner arrays should have only 0s and 1s. Should be of shape (n_samples, n_features, num_classes) or (n_samples * n_features, num_classes).

            threshold (float, optional): Threshold to use to set decoded multilabel values to 0s (< threshold) or 1s (>= threshold). Defaults to 0.5.
        """
        N = 2
        y_pred_proba[y_pred_proba.argsort().argsort() < N] = 0.0
        return y_pred_proba
        # idx = np.argpartition(y_pred_proba.ravel(), k)
        # indices = tuple(
        #     np.array(np.unravel_index(idx, y_pred_proba.shape))[
        #         :, range(min(k, 0), max(k, 0))
        #     ]
        # )

        # y_pred_proba[indices] = 0.0
        # return y_pred_proba

        # return np.where(y_pred_proba >= threshold, 1.0, 0.0)

    @classmethod
    def decode_multiclass(cls, y_pred_proba, reduce_dim=True):
        """Decode probabilities to either one-hot or integer encodings.

        Args:
            y_pred_proba (numpy.ndarray): Probabilities to decode.

            reduce_dim (bool, optional): If True, returns integer encodings of one fewer dimension than ``y_pred_proba``\. Otherwise, returns one-hot encodings where the class with the maximum probability is a 1 and every other class is 0. Defaults to True.

        Returns:
            numpy.ndarray: Integer or one-hot-encoded predictions.
        """
        yt = np.apply_along_axis(cls.mle, axis=-1, arr=y_pred_proba)
        if reduce_dim:
            return np.argmax(yt, axis=-1)
        else:
            return yt

    @classmethod
    def decode_binary_multilab(cls, y_pred):
        """Decode multi-label sigmoid probabilities to integer encodings.

        The predictions should have already undergone sigmoid activation and should be probabilities.

        If sigmoid activation output is >0.5, gets encoded as 1.0; else 0.0. If more than one category is > 0.5, then it is a heterozygote.

        Args:
            y_pred (numpy.ndarray): Model predictions of shape (n_samples * n_features, num_classes) or (n_samples, n_features, num_classes). A threshold should already have been applied to set each class to 0 or 1.

        Returns:
            numpy.ndarray: Integer-decoded multilabel predictions of shape (n_samples * n_features) or (n_samples, n_features).
        """
        y_pred_idx = y_pred.astype(int)
        y_pred_idx = y_pred_idx.astype(str)

        if len(y_pred_idx.shape) < 3:
            y_pred_idx = np.array(
                [
                    "".join(np.atleast_1d(row == "1").nonzero()[0].astype(str))
                    for row in y_pred_idx
                ]
            )
        else:
            y_pred_idx = np.array(
                [
                    "".join(np.atleast_1d(col == "1").nonzero()[0].astype(str))
                    for row in y_pred_idx
                    for col in row
                ]
            )

        try:
            Xt = np.zeros(shape=(y_pred.shape[0], y_pred.shape[1], 4))
        except IndexError:
            Xt = np.zeros(shape=(y_pred.shape[0],))

        mappings = {
            "0": 0,
            "1": 1,
            "2": 2,
            "3": 3,
            "01": 4,
            "02": 5,
            "03": 6,
            "12": 7,
            "13": 8,
            "23": 9,
            "-9": -9,
            "": -9,
        }

        Xt = [mappings[enc] for enc in y_pred_idx]

        if not isinstance(Xt, np.ndarray):
            Xt = np.array(Xt)
        return Xt

    @staticmethod
    def encode_categorical(X):
        """Encode -9 encoded missing values as np.nan.

        Args:
            X (numpy.ndarray): 012-encoded genotypes with -9 as missing values.

        Returns:
            pandas.DataFrame: DataFrame with missing values encoded as np.nan.
        """
        np.nan_to_num(X, copy=False, nan=-9.0)
        X = X.astype(str)
        X[(X == "-9.0") | (X == "-9")] = "none"

        df = pd.DataFrame(X)
        df_incomplete = df.copy()

        # Replace 'none' with np.nan
        for row in df.index:
            for col in df.columns:
                if df_incomplete.iat[row, col] == "none":
                    df_incomplete.iat[row, col] = np.nan

        return df_incomplete

    @staticmethod
    def mle(row):
        """Get the Maximum Likelihood Estimation for the best prediction. Basically, it sets the index of the maxiumum value in a vector (row) to 1.0, since it is one-hot encoded.

        Args:
            row (numpy.ndarray(float)): Row vector with predicted values as floating points.

        Returns:
            numpy.ndarray(float): Row vector with the highest prediction set to 1.0 and the others set to 0.0.
        """
        res = np.zeros(row.shape[0])
        res[np.argmax(row)] = 1
        return res

    @classmethod
    def predict(cls, X, complete_encoded):
        """Evaluate VAE predictions by calculating the highest predicted value.

        Calucalates highest predicted value for each row vector and each class, setting the most likely class to 1.0.

        Args:
            X (numpy.ndarray): Input 012-encoded data.

            complete_encoded (numpy.ndarray): Output one-hot encoded data with the maximum predicted values for each class set to 1.0.

        Returns:
            numpy.ndarray: Imputed one-hot encoded values.

            pandas.DataFrame: One-hot encoded pandas DataFrame with no missing values.
        """

        df = cls.encode_categorical(X)

        # Had to add dropna() to count unique classes while ignoring np.nan
        col_classes = [len(df[c].dropna().unique()) for c in df.columns]
        df_dummies = pd.get_dummies(df)
        mle_complete = None
        for i, cnt in enumerate(col_classes):
            start_idx = int(sum(col_classes[0:i]))
            col_completed = complete_encoded[:, start_idx : start_idx + cnt]
            mle_completed = np.apply_along_axis(
                cls.mle, axis=1, arr=col_completed
            )

            if mle_complete is None:
                mle_complete = mle_completed

            else:
                mle_complete = np.hstack([mle_complete, mle_completed])
        return mle_complete, df_dummies

    def validate_hidden_layers(self, hidden_layer_sizes, num_hidden_layers):
        """Validate hidden_layer_sizes and verify that it is in the correct format.

        Args:
            hidden_layer_sizes (str, int, List[str], or List[int]): Output units for all the hidden layers.

            num_hidden_layers (int): Number of hidden layers to use.

        Returns:
            List[int] or List[str]: List of hidden layer sizes.
        """
        if isinstance(hidden_layer_sizes, (str, int)):
            hidden_layer_sizes = [hidden_layer_sizes] * num_hidden_layers

        # If not all integers
        elif isinstance(hidden_layer_sizes, list):
            if not all(
                [isinstance(x, (str, int)) for x in hidden_layer_sizes]
            ):
                ls = list(set([type(item) for item in hidden_layer_sizes]))
                raise TypeError(
                    f"Variable hidden_layer_sizes must either be None, "
                    f"an integer or string, or a list of integers or "
                    f"strings, but got the following type(s): {ls}"
                )

        else:
            raise TypeError(
                f"Variable hidden_layer_sizes must either be, "
                f"an integer, a string, or a list of integers or strings, "
                f"but got the following type: {type(hidden_layer_sizes)}"
            )

        assert (
            num_hidden_layers == len(hidden_layer_sizes)
            and num_hidden_layers > 0
        ), "num_hidden_layers must be the length of hidden_layer_sizes."

        return hidden_layer_sizes

    def get_hidden_layer_sizes(self, n_dims, n_components, hl_func, vae=False):
        """Get dimensions of hidden layers.

        Args:
            n_dims (int): The number of feature dimensions (columns) (d).

            n_components (int): The number of reduced dimensions (t).

            hl_func (str): The function to use to calculate the hidden layer sizes. Possible options: "midpoint", "sqrt", "log2".

            vae (bool, optional): Whether using the VAE algorithm. If False, then the returned list gets reversed for NLPCA and UBP.

        Returns:
            [int, int, int, ...]: [Number of dimensions in hidden layers].

        Raises:
            ValueError: Too many hidden layers specified. Repeated reduction of layer sizes dips below n_components.
        """
        layers = list()
        if not isinstance(hl_func, list):
            raise TypeError(
                f"hl_func must be of type list, but got {type(hl_func)}."
            )

        units = n_dims
        for func in hl_func:
            if func == "midpoint":
                units = round((units + n_components) / 2)
            elif func == "sqrt":
                units = round(math.sqrt(units))
            elif func == "log2":
                units = round(math.log(units, 2))
            elif isinstance(func, int):
                units = func
            else:
                raise ValueError(
                    f"hidden_layer_sizes must be either integers or any of "
                    f"the following strings: 'midpoint', "
                    f"'sqrt', or 'log2', but got {func} of type {type(func)}"
                )

            if units <= n_components:
                print(
                    f"WARNING: hidden_layer_size reduction became less than n_components. Using only {len(layers)} hidden layers."
                )
                break

            assert units > 0 and units < n_dims, (
                f"The hidden layer sizes must be > 0 and < the number of "
                f"features (i.e., columns) in the dataset, but size was {units}"
            )

            layers.append(units)

        assert (
            layers
        ), "There was an error setting hidden layer sizes. Size list is empty. It is possible that the first 'sqrt' reduction caused units to be <= n_components."

        if not vae:
            layers.reverse()

        return layers

    def validate_model_inputs(self, y, missing_mask, output_shape):
        """Validate inputs to Keras subclass model.

        Args:
            V (numpy.ndarray): Input to refine. Shape: (n_samples, n_components).
            y (numpy.ndarray): Target (but actual input data). Shape: (n_samples, n_features).

            y_test (numpy.ndarray): Target test dataset. Should have been imputed with simple imputer and missing data simulated using SimGenotypeData(). Shape: (n_samples, n_features).

            missing_mask (numpy.ndarray): Missing data mask for y.

            missing_mask_test (numpy.ndarray): Missing data mask for y_test.

            output_shape (int): Output shape for hidden layers.

        Raises:
            TypeError: V, y, missing_mask, output_shape must not be NoneType.
        """
        if y is None:
            raise TypeError("y must not be NoneType.")

        if missing_mask is None:
            raise TypeError("missing_mask must not be NoneType.")

        if output_shape is None:
            raise TypeError("output_shape must not be NoneType.")

    def prepare_training_batches(
        self,
        V,
        y,
        batch_size,
        batch_idx,
        trainable,
        n_components,
        sample_weight,
        missing_mask,
        ubp=True,
    ):
        """Prepare training batches in the custom training loop.

        Args:
            V (numpy.ndarray): Input to batch subset and refine, of shape (n_samples, n_components) (if doing UBP/NLPCA) or (n_samples, n_features) (if doing VAE).

            y (numpy.ndarray): Target to use to refine input V. shape (n_samples, n_features).

            batch_size (int): Batch size to subset.

            batch_idx (int): Current batch index.

            trainable (bool): Whether tensor v should be trainable.

            n_components (int): Number of principal components used in V.

            sample_weight (List[float] or None): List of floats of shape (n_samples,) with sample weights. sample_weight argument must be passed to fit().

            missing_mask (numpy.ndarray): Boolean array with True for missing values and False for observed values.

            ubp (bool, optional): Whether model is UBP/NLPCA (if True) or VAE (if False). Defaults to True.

        Returns:
            tf.Variable: Input tensor v with current batch assigned.
            numpy.ndarray: Current batch of target data (actual input) used to refine v.
            List[float]: Sample weights
            int: Batch starting index.
            int: Batch ending index.
            numpy.ndarray: Batch of y_train target data of shape (batch_size, n_features, n_classes). Only returned for VAE.
        """
        # on_train_batch_begin() method.
        n_samples = y.shape[0]

        # Get current batch size and range.
        # self._batch_idx is set in the UBPCallbacks() callback
        batch_start = batch_idx * batch_size
        batch_end = (batch_idx + 1) * batch_size
        if batch_end > n_samples:
            batch_end = n_samples - 1
            batch_size = batch_end - batch_start

        if ubp:
            # override batches. This model refines the input to fit the output, so
            # v_batch and y_true have to be overridden.
            y_true = y[batch_start:batch_end, :]

            v_batch = V[batch_start:batch_end, :]
            missing_mask_batch = missing_mask[batch_start:batch_end, :]

            if sample_weight is not None:
                sample_weight_batch = sample_weight[batch_start:batch_end, :]
            else:
                sample_weight_batch = None

            v = tf.Variable(
                tf.zeros([batch_size, n_components]),
                trainable=trainable,
                dtype=tf.float32,
            )

            # Assign current batch to tf.Variable v.
            v.assign(v_batch)

            return (
                v,
                y_true,
                sample_weight_batch,
                missing_mask_batch,
                batch_start,
                batch_end,
            )

        else:
            # Using VAE.
            y_true = y[batch_start:batch_end, :]
            v = V[batch_start:batch_end, :]
            missing_mask_batch = missing_mask[batch_start:batch_end, :]

            if sample_weight is not None:
                sample_weight_batch = sample_weight[batch_start:batch_end, :]
            else:
                sample_weight_batch = None

            return (
                y_true,
                sample_weight_batch,
                missing_mask_batch,
            )

    def validate_batch_size(self, X, batch_size):
        """Validate the batch size, and adjust as necessary.

        If the specified batch_size is greater than the number of samples in the input data, it will divide batch_size by 2 until it is less than n_samples.

        Args:
            X (numpy.ndarray): Input data of shape (n_samples, n_features).
            batch_size (int): Batch size to use.

        Returns:
            int: Batch size (adjusted if necessary).
        """
        if batch_size > X.shape[0]:
            while batch_size > X.shape[0]:
                print(
                    "Batch size is larger than the number of samples. "
                    "Dividing batch_size by 2."
                )
                batch_size //= 2
        return batch_size

    def set_compile_params(
        self, optimizer, sample_weights=None, vae=False, act_func="softmax"
    ):
        """Set compile parameters to use.

        Args:
            optimizer (str): Keras optimizer to use. Possible options include: {"adam", "sgd", "adagrad", "adadelta", "adamax", "ftrl", "nadam", "rmsprop"}.

            sample_weights (numpy.ndarray, optional): Sample weight matrix of shape (n_samples, n_features). Defaults to None.

            vae (bool, optional): Whether using the VAE model. Defaults to False.

            act_func (str, optional): Activation function to use. Should be "softmax" if doing multiclass classification, otherwise "sigmoid".

        Returns:
            Dict[str, callable] or Dict[str, Any]: Callables if search_mode is True, otherwise instantiated objects.

        Raises:
            ValueError: Unsupported optimizer specified.
            ValueError: Invalid act_func argument supplied.
        """
        if optimizer.lower() == "adam":
            opt = tf.keras.optimizers.legacy.Adam
        elif optimizer.lower() == "sgd":
            opt = tf.keras.optimizers.legacy.SGD
        elif optimizer.lower() == "adagrad":
            opt = tf.keras.optimizers.legacy.Adagrad
        elif optimizer.lower() == "adadelta":
            opt = tf.keras.optimizers.legacy.Adadelta
        elif optimizer.lower() == "adamax":
            opt = tf.keras.optimizers.legacy.Adamax
        elif optimizer.lower() == "ftrl":
            opt = tf.keras.optimizers.legacy.Ftrl
        elif optimizer.lower() == "nadam":
            opt = tf.keras.optimizers.legacy.Nadam
        elif optimizer.lower() == "rmsprop":
            opt = tf.keras.optimizers.legacy.RMSProp

        if vae:
            if act_func == "softmax":
                loss_func = (
                    NeuralNetworkMethods.make_masked_categorical_crossentropy
                )
            elif act_func == "sigmoid":
                loss_func = (
                    NeuralNetworkMethods.make_masked_binary_crossentropy
                )
            else:
                raise ValueError(
                    f"act_func must be either 'softmax' or 'sigmoid', but got {act_func}"
                )

            loss = loss_func()
            metrics = None

        else:
            # Doing grid search. Params are callables.
            loss = tf.keras.losses.CategoricalCrossentropy(from_logits=True)
            metrics = [tf.keras.metrics.CategoricalAccuracy()]

        return {
            "optimizer": opt,
            "loss": loss,
            "metrics": metrics,
            "run_eagerly": False,
        }

    @staticmethod
    def init_weights(dim1, dim2, w_mean=0, w_stddev=0.01):
        """Initialize random weights to use with the model.

        Args:
            dim1 (int): Size of first dimension.

            dim2 (int): Size of second dimension.

            w_mean (float, optional): Mean of normal distribution. Defaults to 0.

            w_stddev (float, optional): Standard deviation of normal distribution. Defaults to 0.01.
        """
        # Get reduced-dimension dataset.
        return np.random.normal(loc=w_mean, scale=w_stddev, size=(dim1, dim2))

    def reset_seeds(self):
        """Reset random seeds for initializing weights."""
        seed1 = np.random.randint(1, 1e6)
        seed2 = np.random.randint(1, 1e6)
        seed3 = np.random.randint(1, 1e6)
        np.random.seed(seed1)
        random.seed(seed2)
        if tf.__version__[0] == "2":
            tf.random.set_seed(seed3)
        else:
            tf.set_random_seed(seed3)

    @staticmethod
    def masked_mse(self, X_true, X_pred, mask):
        """Calculates mean squared error with missing values ignored.

        Args:
            X_true (numpy.ndarray): One-hot encoded input data.
            X_pred (numpy.ndarray): Predicted values.
            mask (numpy.ndarray): One-hot encoded missing data mask.

        Returns:
            float: Mean squared error calculation.
        """
        return np.square(np.subtract(X_true[mask], X_pred[mask])).mean()

    @staticmethod
    def make_masked_binary_accuracy(class_weight=None, is_vae=True):
        """Make binary accuracy metric with missing mask.

        Args:
            class_weight (Dict[int, float], optional): Class weights to reduce class imbalance. Defaults to None.

            is_vae (bool, optional): Whether model is VAE or not. Defaults to True.

        Returns:
            callable: Function that calculates categorical crossentropy loss.
        """

        @tf.function
        def masked_binary_accuracy(y_true, y_pred, sample_weight=None):
            """Custom neural network metric function with missing mask.

            Ignores missing data in the calculation of the loss function.

            Args:
                y_true (tensorflow.Tensor): Input multilabel encoded 3D tensor.
                y_pred (tensorflow.Tensor): Predicted values from model.
                sample_weight (numpy.ndarray): 2D matrix of sample weights.

            Returns:
                float: Binary accuracy calculated with missing data masked.
            """
            return ba(
                y_true,
                y_pred,
                sample_weight=sample_weight,
            )

        return masked_binary_accuracy

    @staticmethod
    def make_masked_binary_crossentropy(class_weight=None, is_vae=True):
        """Make binary crossentropy loss function with missing mask.

        Args:
            class_weight (Dict[int, float], optional): Class weights to reduce class imbalance. Defaults to None.

            is_vae (bool, optional): Whether model is VAE or not. Defaults to True.

        Returns:
            callable: Function that calculates categorical crossentropy loss.
        """

        @tf.function
        def masked_binary_crossentropy(y_true, y_pred, sample_weight=None):
            """Custom loss function for with missing mask applied.

            Ignores missing data in the calculation of the loss function.

            Args:
                y_true (tensorflow.tensor): Input one-hot encoded 3D tensor.

                y_pred (tensorflow.tensor): Predicted values, should have undergone sigmoid activation.

                sample_weight (numpy.ndarray): 2D matrix of sample weights.

            Returns:
                float: Binary crossentropy loss value.
            """
            return bce(
                y_true,
                y_pred,
                sample_weight=sample_weight,
            )

        return masked_binary_crossentropy

    @staticmethod
    def make_masked_categorical_accuracy():
        """Make categorical crossentropy loss function with missing mask.

        Args:
            class_weight (Dict[int, float): Weights for each class.
            is_vae (bool, optional): Whether using VAE model. Defaults to False.

        Returns:
            callable: Function that calculates categorical crossentropy loss.
        """

        @tf.function
        def masked_categorical_accuracy(y_true, y_pred, sample_weight=None):
            """Custom loss function for neural network model with missing mask.
            Ignores missing data in the calculation of the loss function.
            Args:
                y_true (tensorflow.tensor): Input one-hot encoded 3D tensor.
                y_pred (tensorflow.tensor): Predicted values.
                sample_weight (numpy.ndarray): 2D matrix of sample weights.

            Returns:
                float: Mean squared error loss value with missing data masked.
            """
            # # Mask out missing values.
            # y_true_masked = tf.boolean_mask(
            #     y_true,
            #     tf.reduce_any(tf.not_equal(y_true, -1), axis=-1),
            # )

            # y_pred_masked = tf.boolean_mask(
            #     y_pred,
            #     tf.reduce_any(tf.not_equal(y_true, -1), axis=-1),
            # )

            return cca(
                y_true,
                y_pred,
                sample_weight=sample_weight,
            )

        return masked_categorical_accuracy

    @staticmethod
    def make_masked_categorical_crossentropy():
        """Make categorical crossentropy loss function with missing mask.

        Returns:
            callable: Function that calculates categorical crossentropy loss.
        """

        @tf.function
        def masked_categorical_crossentropy(
            y_true, y_pred, sample_weight=None
        ):
            """Custom loss function for neural network model with missing mask.
            Ignores missing data in the calculation of the loss function.

            Args:
                y_true (tensorflow.tensor): Input one-hot encoded 3D tensor.
                y_pred (tensorflow.tensor): Predicted values.
                sample_weight (numpy.ndarray): 2D matrix of sample weights.

            Returns:
                float: Mean squared error loss value with missing data masked.
            """
            # Mask out missing values.
            # y_true_masked = tf.boolean_mask(
            #     y_true,
            #     tf.reduce_any(tf.not_equal(y_true, -1), axis=-1),
            # )

            # y_pred_masked = tf.boolean_mask(
            #     y_pred,
            #     tf.reduce_any(tf.not_equal(y_true, -1), axis=-1),
            # )

            return cce(
                y_true,
                y_pred,
                sample_weight=sample_weight,
            )

        return masked_categorical_crossentropy

    @staticmethod
    def kl_divergence(z_mean, z_log_var, kl_weight=0.5):
        kl_loss = -0.5 * (
            1 + z_log_var - tf.square(z_mean) - tf.exp(z_log_var)
        )
        return tf.reduce_mean(tf.reduce_sum(kl_loss, axis=-1))

        # Another way of doing it.
        # TODO: Test both ways.
        # z_sigma = tf.math.exp(0.5 * z_log_var)
        # return tf.reduce_sum(
        #         tf.math.square(z_mean) + tf.math.square(z_sigma) - z_log_var - 1.0,
        #         axis=-1,
        #     )

    def make_reconstruction_loss(self):
        """Make loss function for use with a keras model.

        Returns:
            callable: Function that calculates loss.
        """

        def reconstruction_loss(input_and_mask, y_pred):
            """Custom loss function for neural network model with missing mask.

            Ignores missing data in the calculation of the loss function.

            Args:
                input_and_mask (numpy.ndarray): Input one-hot encoded array with missing values also one-hot encoded and h-stacked.

                y_pred (numpy.ndarray): Predicted values.

            Returns:
                float: Mean squared error loss value with missing data masked.
            """
            n_features = y_pred.numpy().shape[1]

            true_indices = range(n_features)
            missing_indices = range(n_features, n_features * 2)

            # Split features and missing mask.
            y_true = tf.gather(input_and_mask, true_indices, axis=1)
            missing_mask = tf.gather(input_and_mask, missing_indices, axis=1)

            observed_mask = tf.subtract(1.0, missing_mask)
            y_true_observed = tf.multiply(y_true, observed_mask)
            pred_observed = tf.multiply(y_pred, observed_mask)

            # loss_fn = tf.keras.losses.CategoricalCrossentropy()
            # return loss_fn(y_true_observed, pred_observed)

            return tf.keras.metrics.mean_squared_error(
                y_true=y_true_observed, y_pred=pred_observed
            )

        return reconstruction_loss

    @staticmethod
    def normalize_data(data):
        """Normalize data between 0 and 1."""
        return (data - np.min(data)) / (np.max(data) - np.min(data))

    @staticmethod
    def normalize_sum_to_1(d, target=1.0):
        factor = target / sum(d.values())
        return {k: v * factor for k, v in d.items()}

    @staticmethod
    def smooth_weights(d, mu=0.15):
        total = np.sum(list(d.values()))
        keys = d.keys()
        class_weight = dict()

        for k in keys:
            score = math.log(mu * total / float(d[k]))
            class_weight[k] = score if score > 1.0 else 1.0

        return class_weight

    @classmethod
    def get_class_weights(
        cls,
        y_true,
        original_missing_mask,
        user_weights=None,
        return_1d=False,
        method="auto",
    ):
        """Get class weights for each column in a 2D matrix.

        Args:
            y_true (numpy.ndarray): True target values.

            original_missing_mask (numpy.ndarray): Boolean mask with missing values set to True and non-missing to False.

            user_weights (Dict[int, float], optional): Class weights if user-provided.

            return_1d (bool, optional): If True, returns a dictionary of class weights, with integer encodings as keys and the corresponding class weights as keys. If False, returns 2D sample_weight matrix. Defaults to False.

        Returns:
            numpy.ndarray or Dict[int, float]: Sample weights per column of shape (n_samples, n_features) if return_1d is False. Dictionary of class weights if True.
        """
        # Get list of class_weights (per-column).
        class_weights = list()
        sample_weight = np.zeros(y_true.shape)
        if user_weights is not None:
            # Set user-defined sample_weights
            for k in user_weights.keys():
                sample_weight[y_true == k] = user_weights[k]

        elif return_1d:
            y_true_1d = y_true.flatten()

            if method == "auto":
                sample_weight = dict(
                    zip(
                        np.unique(y_true_1d),
                        compute_class_weight(
                            "balanced",
                            classes=np.unique(y_true_1d),
                            y=y_true_1d,
                        ),
                    )
                )

            elif method == "logsmooth":
                counts = np.unique(y_true_1d, return_counts=True)
                sample_weight = dict(zip(counts[0], counts[1]))
                sample_weight.pop(-9)
                sample_weight = cls.smooth_weights(sample_weight)
                sample_weight[-9] = 0.0

        else:
            # Automatically get class weights to set sample_weight.
            for i in np.arange(y_true.shape[1]):
                mm = ~original_missing_mask[:, i]
                classes = np.unique(y_true[mm, i])
                cw = compute_class_weight(
                    "balanced",
                    classes=classes,
                    y=y_true[mm, i],
                )

                class_weights.append({k: v for k, v in zip(classes, cw)})

            # Make sample_weight_matrix from automatic per-column class_weights.
            for i, w in enumerate(class_weights):
                for j in range(3):
                    if j in w:
                        sample_weight[y_true[:, i] == j, i] = w[j]

        return sample_weight

    @staticmethod
    def write_gt_state_probs(
        y_pred,
        y_pred_1d,
        y_true,
        y_true_1d,
        nn_method,
        sim_missing_mask,
        original_missing_mask,
        prefix="imputer",
    ):
        bin_mapping = np.array(
            [np.array2string(x) for row in y_pred for x in row]
        )

        bin_mapping = np.reshape(bin_mapping, y_pred_1d.shape)

        y_true_2d = np.reshape(y_true_1d, y_true.shape)
        bin_mapping_2d = np.reshape(bin_mapping, y_true.shape)
        y_pred_2d = np.reshape(y_pred_1d, y_true.shape)

        include = np.logical_and(sim_missing_mask, ~original_missing_mask)

        gt_dist = list()
        colors = []
        for yt, yp, ypd, mask in zip(
            y_true_2d,
            bin_mapping_2d,
            y_pred_2d,
            include,
        ):
            sites = dict()
            row_colors = []
            for i, (yt_site, mask_site) in enumerate(zip(yt, mask)):
                if mask_site:
                    sites[
                        f"Site Index {i},Probability Vector,Imputed Genotype,Expected Genotype"
                    ] = f"{i},{yp[i]},{ypd[i]},{yt_site}"
                    if ypd[i] == yt_site:
                        row_colors.append("blue")
                    else:
                        sites[
                            f"Site Index {i},Probability Vector,Imputed Genotype,Expected Genotype"
                        ] = f"{i},{yp[i]},{ypd[i]},{yt_site}"
                        row_colors.append("orange")
                else:
                    sites[
                        f"Site Index {i},Probability Vector,Imputed Genotype,Expected Genotype"
                    ] = f"{i},{np.array2string(np.array([0.0, 0.0, 0.0]))},0,0"
                    row_colors.append("gray")
            gt_dist.append(sites)
            colors.append(row_colors)

        gt_df = pd.DataFrame.from_records(gt_dist)
        gt_df.to_csv(
            os.path.join(
                f"{prefix}_output",
                "logs",
                "Unsupervised",
                nn_method,
                "genotype_state_proba.csv",
            ),
            index=False,
            header=False,
        )

        # Reload the data

        data = pd.read_csv(
            os.path.join(
                f"{prefix}_output",
                "logs",
                "Unsupervised",
                nn_method,
                "genotype_state_proba.csv",
            ),
            header=None,
        )

        # Parse the original data into separate dataframes for imputedGT and expectedGT
        imputedGT_data = data.applymap(lambda x: int(x.split(",")[2]))
        expectedGT_data = data.applymap(lambda x: int(x.split(",")[3]))

        # Determine the binary mask based on whether imputedGT and expectedGT are the same

        mask = imputedGT_data == expectedGT_data

        # Create a new figure and set its size
        plt.figure(figsize=(12, 6))

        from matplotlib.colors import ListedColormap

        rgb_colors = sns.color_palette(
            [color for sublist in colors for color in sublist]
        )
        cmap = ListedColormap(rgb_colors)

        # Create a heatmap
        sns.heatmap(mask, cmap=cmap, cbar=False)

        # Set the title and labels
        plt.title("Expected Genotypes for Simulated Genotypes")
        plt.xlabel("Column Index")
        plt.ylabel("Row Index")

        # Create a custom legend
        import matplotlib.patches as mpatches

        green_patch = mpatches.Patch(color="blue", label="Agreement")
        orange_patch = mpatches.Patch(color="orange", label="Disagreement")
        gray_patch = mpatches.Patch(color="gray", label="Not Simulated")

        plt.legend(
            handles=[green_patch, orange_patch, gray_patch], loc="lower right"
        )

        outfile = os.path.join(
            f"{prefix}_output",
            "plots",
            "Unsupervised",
            nn_method,
            "gt_state_proba.png",
        )

        plt.savefig(outfile, bbox_inches="tight", facecolor="white")

    # @staticmethod
    # def write_gt_state_probs(
    #     y_pred,
    #     y_pred_1d,
    #     y_true,
    #     y_true_1d,
    #     nn_method,
    #     sim_missing_mask,
    #     original_missing_mask,
    #     prefix="imputer",
    # ):
    #     bin_mapping = np.array(
    #         [np.array2string(x) for row in y_pred for x in row]
    #     )

    #     bin_mapping = np.reshape(bin_mapping, y_pred_1d.shape)

    #     y_true_2d = np.reshape(y_true_1d, y_true.shape)
    #     bin_mapping_2d = np.reshape(bin_mapping, y_true.shape)
    #     y_pred_2d = np.reshape(y_pred_1d, y_true.shape)

    #     gt_dist = list()
    #     for yt, yp, ypd in zip(y_true_2d, bin_mapping_2d, y_pred_2d):
    #         sites = dict()
    #         for i, yt_site in enumerate(yt):
    #             sites[
    #                 f"Site Index {i},Probability Vector,Imputed Genotype,Expected Genotype"
    #             ] = f"{i},{yp[i]},{ypd[i]},{yt_site}"
    #         gt_dist.append(sites)

    #     gt_df = pd.DataFrame.from_records(gt_dist)
    #     gt_df.to_csv(
    #         os.path.join(
    #             f"{prefix}_output",
    #             "logs",
    #             "Unsupervised",
    #             nn_method,
    #             "genotype_state_proba.csv",
    #         ),
    #         index=False,
    #         header=False,
    #     )



import logging
import os
import sys
import warnings

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # or any {'0', '1', '2', '3'}
logging.getLogger("tensorflow").disabled = True

# Import tensorflow with reduced warnings.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

# noinspection PyPackageRequirements
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns of GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated

from tensorflow.keras.layers import (
    Dropout,
    Dense,
    Reshape,
    LeakyReLU,
    PReLU,
)

from tensorflow.keras.regularizers import l1_l2

# Custom Modules
try:
    from ..neural_network_methods import NeuralNetworkMethods
except (ModuleNotFoundError, ValueError, ImportError):
    from impute.unsupervised.neural_network_methods import NeuralNetworkMethods


class UBPPhase1(tf.keras.Model):
    """UBP Phase 1 single layer perceptron model to train predict imputations.

    This model is subclassed from the tensorflow/ Keras framework.

    UBPPhase1 subclasses the tf.keras.Model and overrides the train_step function, which does training and evalutation for each batch in each epoch.

    UBPPhase1 is a single-layer perceptron model used to initially refine V. After Phase 1 the Phase 1 weights are discarded.

    Args:
        V (numpy.ndarray(float)): V should have been randomly initialized and will be used as the input data that gets refined during training. Defaults to None.

        y (numpy.ndarray): Target values to predict. Actual input data. Defaults to None.

        batch_size (int, optional): Batch size per epoch. Defaults to 32.

        missing_mask (numpy.ndarray): Missing data mask for y. Defaults to None.

        output_shape (int): Output units for n_features dimension. Output will be of shape (batch_size, n_features). Defaults to None.

        n_components (int, optional): Number of features in input V to use. Defaults to 3.

        weights_initializer (str, optional): Kernel initializer to use for initializing model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (NoneType, optional): Output units for each hidden layer. List should be of same length as the number of hidden layers. Not used for UBP Phase 1, but is here for compatibility. Defaults to "midpoint".

        num_hidden_layers (int, optional): Number of hidden layers to use. Not used in UBP Phase 1, but is here for compatibility. Defaults to 1.

        hidden_activation (str, optional): Activation function to use for hidden layers. Defaults to "elu".

        l1_penalty (float, optional): L1 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate during training to reduce overfitting. Must be a float between 0 and 1. Defaults to 0.2.

        num_classes (int, optional): Number of classes in output. Corresponds to the 3rd dimension of the output shape (batch_size, n_features, num_classes). Defaults to 3.

        phase (int, optional): Current phase if doing UBP model. Defaults to 1.

        sample_weight (numpy.ndarray, optional): 2D sample weights of shape (n_samples, n_features). Should have values for each class weighted. Defaults to None.

    Example:
        >>>model = UBPPhase1(V=V, y=y, batch_size=32, missing_mask=missing_mask, output_shape=y_train.shape[1], n_components=3, weights_initializer="glorot_normal", hidden_layer_sizes="midpoint", num_hidden_layers=1, hidden_activation="elu", l1_penalty=1e-6, l2_penalty=1e-6, num_classes=3, phase=3)
        >>>model.compile(optimizer=optimizer, loss=loss_func, metrics=[my_metrics], run_eagerly=True)
        >>>history = model.fit(X, y, batch_size=batch_size, epochs=epochs, callbacks=[MyCallback()], validation_split=validation_split, shuffle=False)

    Raises:
        TypeError: V, y, missing_mask, output_shape must not be NoneType.
        ValueError: Maximum of 5 hidden layers.
    """

    def __init__(
        self,
        V=None,
        y=None,
        batch_size=32,
        missing_mask=None,
        output_shape=None,
        n_components=3,
        weights_initializer="glorot_normal",
        hidden_layer_sizes="midpoint",
        num_hidden_layers=1,
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        num_classes=3,
        phase=1,
        sample_weight=None,
    ):
        super(UBPPhase1, self).__init__()

        nn = NeuralNetworkMethods()
        self.nn = nn

        if V is None:
            self._V = nn.init_weights(y.shape[0], n_components)
        elif isinstance(V, dict):
            self._V = V[n_components]
        else:
            self._V = V

        self._y = y

        hidden_layer_sizes = nn.validate_hidden_layers(
            hidden_layer_sizes, num_hidden_layers
        )

        hidden_layer_sizes = nn.get_hidden_layer_sizes(
            y.shape[1], self._V.shape[1], hidden_layer_sizes
        )

        nn.validate_model_inputs(y, missing_mask, output_shape)

        self._missing_mask = missing_mask
        self.weights_initializer = weights_initializer
        self.phase = phase
        self.dropout_rate = dropout_rate
        self._sample_weight = sample_weight

        ### NOTE: I tried using just _V as the input to be refined, but it
        # wasn't getting updated. So I copy it here and it works.
        # V_latent is refined during train_step.
        self.V_latent_ = self._V.copy()

        # Initialize parameters used during train_step() and test_step.
        # input_with_mask_ is set during the UBPCallbacks() execution.
        self._batch_idx = 0
        self._batch_size = batch_size
        self.n_components = n_components

        if l1_penalty == 0.0 and l2_penalty == 0.0:
            kernel_regularizer = None
        else:
            kernel_regularizer = l1_l2(l1_penalty, l2_penalty)

        self.kernel_regularizer = kernel_regularizer
        kernel_initializer = weights_initializer

        # Construct single-layer perceptron.

        self.dense1 = Dense(
            output_shape * num_classes,
            input_shape=(n_components,),
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
        )

        self.rshp = Reshape((output_shape, num_classes))

    def call(self, inputs):
        x = self.dense1(inputs)
        return self.rshp(x)

    def model(self):
        x = tf.keras.Input(shape=(self.n_components,))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def set_model_outputs(self):
        x = tf.keras.Input(shape=(self.n_components,))
        model = tf.keras.Model(inputs=[x], outputs=self.call(x))
        self.outputs = model.outputs

    def train_step(self, data):
        """Train step function. Parameters are set in the UBPCallbacks callback"""
        y = self._y

        (
            v,
            y_true,
            sample_weight,
            missing_mask,
            batch_start,
            batch_end,
        ) = self.nn.prepare_training_batches(
            self.V_latent_,
            y,
            self._batch_size,
            self._batch_idx,
            True,
            self.n_components,
            self._sample_weight,
            self._missing_mask,
        )

        src = [v]

        if sample_weight is not None:
            sample_weight_masked = tf.convert_to_tensor(
                sample_weight[~missing_mask], dtype=tf.float32
            )
        else:
            sample_weight_masked = None

        y_true_masked = tf.boolean_mask(
            tf.convert_to_tensor(y_true, dtype=tf.float32),
            tf.reduce_any(tf.not_equal(y_true, -1), axis=2),
        )

        # NOTE: Earlier model architectures incorrectly
        # applied one gradient to all the variables, including
        # the weights and v. Here we apply them separately, per
        # the UBP manuscript.
        with tf.GradientTape(persistent=True) as tape:
            # Forward pass. Watch input tensor v.
            tape.watch(v)
            y_pred = self(v, training=True)
            y_pred_masked = tf.boolean_mask(
                y_pred, tf.reduce_any(tf.not_equal(y_true, -1), axis=2)
            )
            ### NOTE: If you get the error, "'tuple' object has no attribute
            ### 'rank'", then convert y_true to a tensor object."
            loss = self.compiled_loss(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
                regularization_losses=self.losses,
            )

        # Refine the watched variables with
        # gradient descent backpropagation
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(
            zip(gradients, self.trainable_variables)
        )

        # Apply separate gradients to v.
        vgrad = tape.gradient(loss, src)
        self.optimizer.apply_gradients(zip(vgrad, src))

        del tape

        ### NOTE: If you get the error, "'tuple' object has no attribute
        ### 'rank', then convert y_true to a tensor object."
        self.compiled_metrics.update_state(
            y_true_masked,
            y_pred_masked,
            sample_weight=sample_weight_masked,
        )

        # NOTE: run_eagerly must be set to True in the compile() method for this
        # to work. Otherwise it can't convert a Tensor object to a numpy array.
        # There is really no other way to set v back to V_latent_ in graph
        # mode as far as I know. eager execution is slower, so it would be nice
        # to find a way to do this without converting to numpy.
        self.V_latent_[batch_start:batch_end, :] = v.numpy()

        # history object that gets returned from model.fit().
        return {m.name: m.result() for m in self.metrics}

    @property
    def V_latent(self):
        """Randomly initialized input that gets refined during training.
        :noindex:
        """
        return self.V_latent_

    @property
    def batch_size(self):
        """Batch (=step) size per epoch.
        :noindex:
        """
        return self._batch_size

    @property
    def batch_idx(self):
        """Current batch (=step) index.
        :noindex:
        """
        return self._batch_idx

    @property
    def y(self):
        """Input dataset.
        :noindex:
        """
        return self._y

    @property
    def missing_mask(self):
        """Missing mask of shape (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._missing_mask

    @property
    def sample_weight(self):
        """Sample weights of shape (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._sample_weight

    @V_latent.setter
    def V_latent(self, value):
        """Set randomly initialized input. Gets refined during training.
        :noindex:
        """
        self.V_latent_ = value

    @batch_size.setter
    def batch_size(self, value):
        """Set batch_size parameter.
        :noindex:
        """
        self._batch_size = int(value)

    @batch_idx.setter
    def batch_idx(self, value):
        """Set current batch (=step) index.
        :noindex:
        """
        self._batch_idx = int(value)

    @y.setter
    def y(self, value):
        """Set y after each epoch.
        :noindex:
        """
        self._y = value

    @missing_mask.setter
    def missing_mask(self, value):
        """Set missing_mask after each epoch.
        :noindex:
        """
        self._missing_mask = value

    @sample_weight.setter
    def sample_weight(self, value):
        """Set sample_weight after each epoch.
        :noindex:
        """
        self._sample_weight = value


class UBPPhase2(tf.keras.Model):
    """UBP Phase 2 model to train and use to predict imputations.

    UBPPhase2 subclasses the tf.keras.Model and overrides the train_step function, which does training for each batch in each epoch.

    Phase 2 does not refine V, it just refines the weights.

    Args:
        V (numpy.ndarray(float)): V should have been randomly initialized and will be used as the input data that gets refined during training. Defaults to None.

        y (numpy.ndarray): Target values to predict. Actual input data. Defaults to None.

        batch_size (int, optional): Batch size per epoch. Defaults to 32.

        missing_mask (numpy.ndarray): Missing data mask for y. Defaults to None.

        output_shape (int): Output units for n_features dimension. Output will be of shape (batch_size, n_features). Defaults to None.

        n_components (int, optional): Number of features in input V to use. Defaults to 3.

        weights_initializer (str, optional): Kernel initializer to use for initializing model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (NoneType, optional): Output units for each hidden layer. List should be of same length as the number of hidden layers. Not used for UBP Phase 1, but is here for compatibility. Defaults to "midpoint".

        num_hidden_layers (int, optional): Number of hidden layers to use. Not used in UBP Phase 1, but is here for compatibility. Defaults to 1.

        hidden_activation (str, optional): Activation function to use for hidden layers. Defaults to "elu".

        l1_penalty (float, optional): L1 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate during training to reduce overfitting. Must be a float between 0 and 1. Defaults to 0.2.

        num_classes (int, optional): Number of classes in output. Corresponds to the 3rd dimension of the output shape (batch_size, n_features, num_classes). Defaults to 3.

        phase (int, optional): Current phase if doing UBP model. Defaults to 1.

        sample_weight (numpy.ndarray, optional): 2D sample weights of shape (n_samples, n_features). Should have values for each class weighted. Defaults to None.

    Example:
        >>>model = UBPPhase2(V=V, y=y, batch_size=32, missing_mask=missing_mask, output_shape=y_train.shape[1], n_components=3, weights_initializer="glorot_normal", hidden_layer_sizes="midpoint", num_hidden_layers=1, hidden_activation="elu", l1_penalty=1e-6, l2_penalty=1e-6, num_classes=3, phase=3)
        >>>
        >>>model.compile(optimizer=optimizer, loss=loss_func, metrics=[my_metrics], run_eagerly=True)
        >>>
        >>>history = model.fit(X, y, batch_size=batch_size, epochs=epochs, callbacks=[MyCallback()], validation_split=validation_split, shuffle=False)

    Raises:
        TypeError: V, y, missing_mask, output_shape must not be NoneType.
        ValueError: Maximum of 5 hidden layers.
    """

    def __init__(
        self,
        V=None,
        y=None,
        batch_size=32,
        missing_mask=None,
        output_shape=None,
        n_components=3,
        weights_initializer="glorot_normal",
        hidden_layer_sizes="midpoint",
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        num_classes=3,
        phase=2,
        sample_weight=None,
    ):
        super(UBPPhase2, self).__init__()

        nn = NeuralNetworkMethods()
        self.nn = nn

        if V is None:
            self._V = nn.init_weights(y.shape[0], n_components)
        elif isinstance(V, dict):
            self._V = V[n_components]
        else:
            self._V = V

        self._y = y

        hidden_layer_sizes = nn.validate_hidden_layers(
            hidden_layer_sizes, num_hidden_layers
        )

        hidden_layer_sizes = nn.get_hidden_layer_sizes(
            y.shape[1], self._V.shape[1], hidden_layer_sizes
        )

        nn.validate_model_inputs(y, missing_mask, output_shape)

        self._missing_mask = missing_mask
        self.weights_initializer = weights_initializer
        self.phase = phase
        self.dropout_rate = dropout_rate
        self._sample_weight = sample_weight

        ### NOTE: I tried using just _V as the input to be refined, but it
        # wasn't getting updated. So I copy it here and it works.
        # V_latent is refined during train_step.
        self.V_latent_ = self._V.copy()

        # Initialize parameters used during train_step.
        self._batch_idx = 0
        self._batch_size = batch_size
        self.n_components = n_components

        if l1_penalty == 0.0 and l2_penalty == 0.0:
            kernel_regularizer = None
        else:
            kernel_regularizer = l1_l2(l1_penalty, l2_penalty)

        self.kernel_regularizer = kernel_regularizer
        kernel_initializer = weights_initializer

        if hidden_activation.lower() == "leaky_relu":
            activation = LeakyReLU(alpha=0.01)

        elif hidden_activation.lower() == "prelu":
            activation = PReLU()

        elif hidden_activation.lower() == "selu":
            activation = "selu"
            kernel_initializer = "lecun_normal"

        else:
            activation = hidden_activation

        if num_hidden_layers > 5:
            raise ValueError(
                f"The maximum number of hidden layers is 5, but got "
                f"{num_hidden_layers}"
            )

        self.dense2 = None
        self.dense3 = None
        self.dense4 = None
        self.dense5 = None

        # Construct multi-layer perceptron.
        # Add hidden layers dynamically.
        self.dense1 = Dense(
            hidden_layer_sizes[0],
            input_shape=(n_components,),
            activation=activation,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
        )

        if num_hidden_layers >= 2:
            self.dense2 = Dense(
                hidden_layer_sizes[1],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers >= 3:
            self.dense3 = Dense(
                hidden_layer_sizes[2],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers >= 4:
            self.dense4 = Dense(
                hidden_layer_sizes[3],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers == 5:
            self.dense5 = Dense(
                hidden_layer_sizes[4],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        self.output1 = Dense(
            output_shape * num_classes,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
        )

        self.rshp = Reshape((output_shape, num_classes))

        self.dropout_layer = Dropout(rate=dropout_rate)

    def call(self, inputs, training=None):
        x = self.dense1(inputs)
        x = self.dropout_layer(x, training=training)
        if self.dense2 is not None:
            x = self.dense2(x)
            x = self.dropout_layer(x, training=training)
        if self.dense3 is not None:
            x = self.dense3(x)
            x = self.dropout_layer(x, training=training)
        if self.dense4 is not None:
            x = self.dense4(x)
            x = self.dropout_layer(x, training=training)
        if self.dense5 is not None:
            x = self.dense5(x)
            x = self.dropout_layer(x, training=training)

        x = self.output1(x)
        return self.rshp(x)

    def model(self):
        x = tf.keras.Input(shape=(self.n_components,))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def set_model_outputs(self):
        x = tf.keras.Input(shape=(self.n_components,))
        model = tf.keras.Model(inputs=[x], outputs=self.call(x))
        self.outputs = model.outputs

    def train_step(self, data):
        """Train step function. Parameters are set in the UBPCallbacks callback"""
        y = self._y

        (
            v,
            y_true,
            sample_weight,
            missing_mask,
            _,
            __,
        ) = self.nn.prepare_training_batches(
            self.V_latent_,
            y,
            self._batch_size,
            self._batch_idx,
            True,
            self.n_components,
            self._sample_weight,
            self._missing_mask,
        )

        if sample_weight is not None:
            sample_weight_masked = tf.convert_to_tensor(
                sample_weight[~missing_mask], dtype=tf.float32
            )
        else:
            sample_weight_masked = None

        y_true_masked = tf.boolean_mask(
            tf.convert_to_tensor(y_true, dtype=tf.float32),
            tf.reduce_any(tf.not_equal(y_true, -1), axis=2),
        )

        # NOTE: Earlier model architectures incorrectly
        # applied one gradient to all the variables, including
        # the weights and v. Here we apply them separately, per
        # the UBP manuscript.
        with tf.GradientTape() as tape:
            # Forward pass
            y_pred = self(v, training=True)
            y_pred_masked = tf.boolean_mask(
                y_pred, tf.reduce_any(tf.not_equal(y_true, -1), axis=2)
            )
            ### NOTE: If you get the error, "'tuple' object has no attribute
            ### 'rank'", then convert y_true to a tensor object."
            loss = self.compiled_loss(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
                regularization_losses=self.losses,
            )

        # Refine the watched variables with backpropagation
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(
            zip(gradients, self.trainable_variables)
        )

        ### NOTE: If you get the error, "'tuple' object has no attribute
        ### 'rank', then convert y_true to a tensor object."
        self.compiled_metrics.update_state(
            y_true_masked,
            y_pred_masked,
            sample_weight=sample_weight_masked,
        )

        # history object that gets returned from fit().
        return {m.name: m.result() for m in self.metrics}

    @property
    def V_latent(self):
        """Randomly initialized input variable that gets refined during training.
        :noindex:
        """
        return self.V_latent_

    @property
    def batch_size(self):
        """Batch (=step) size per epoch.
        :noindex:
        """
        return self._batch_size

    @property
    def batch_idx(self):
        """Current batch (=step) index.
        :noindex:
        """
        return self._batch_idx

    @property
    def y(self):
        """Full input dataset.
        :noindex:
        """
        return self._y

    @property
    def missing_mask(self):
        """Get missing_mask for current epoch.
        :noindex:
        """
        return self._missing_mask

    @property
    def sample_weight(self):
        """Get sample_weight for current epoch.
        :noindex:
        """
        return self._sample_weight

    @V_latent.setter
    def V_latent(self, value):
        """Set randomly initialized input variable. Gets refined during training.
        :noindex:
        """
        self.V_latent_ = value

    @batch_size.setter
    def batch_size(self, value):
        """Set batch_size parameter.
        :noindex:
        """
        self._batch_size = int(value)

    @batch_idx.setter
    def batch_idx(self, value):
        """Set current batch (=step) index.
        :noindex:
        """
        self._batch_idx = int(value)

    @y.setter
    def y(self, value):
        """Set y after each epoch.
        :noindex:
        """
        self._y = value

    @missing_mask.setter
    def missing_mask(self, value):
        """Set missing_mask after each epoch.
        :noindex:
        """
        self._missing_mask = value

    @sample_weight.setter
    def sample_weight(self, value):
        """Set sample_weight after each epoch.
        :noindex:
        """
        self._sample_weight = value


class UBPPhase3(tf.keras.Model):
    """UBP Phase 3 model to train and use to predict imputations.

    UBPPhase3 subclasses the tf.keras.Model and overrides the train_step function, which does training and evaluation for each batch in each single epoch.

    Phase 3 Refines both the weights and V.

    Args:
        V (numpy.ndarray(float)): V should have been randomly initialized and will be used as the input data that gets refined during training. Defaults to None.

        y (numpy.ndarray): Target values to predict. Actual input data. Defaults to None.

        batch_size (int, optional): Batch size per epoch. Defaults to 32.

        missing_mask (numpy.ndarray): Missing data mask for y. Defaults to None.

        output_shape (int): Output units for n_features dimension. Output will be of shape (batch_size, n_features). Defaults to None.

        n_components (int, optional): Number of features in input V to use. Defaults to 3.

        weights_initializer (str, optional): Kernel initializer to use for initializing model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (NoneType, optional): Output units for each hidden layer. List should be of same length as the number of hidden layers. Not used for UBP Phase 1, but is here for compatibility. Defaults to "midpoint".

        num_hidden_layers (int, optional): Number of hidden layers to use. Not used in UBP Phase 1, but is here for compatibility. Defaults to 1.

        hidden_activation (str, optional): Activation function to use for hidden layers. Defaults to "elu".

        l1_penalty (float, optional): L1 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate during training to reduce overfitting. Must be a float between 0 and 1. Defaults to 0.2.

        num_classes (int, optional): Number of classes in output. Corresponds to the 3rd dimension of the output shape (batch_size, n_features, num_classes). Defaults to 3.

        phase (int, optional): Current phase if doing UBP model. Defaults to 1.

        sample_weight (numpy.ndarray, optional): 2D sample weights of shape (n_samples, n_features). Should have values for each class weighted. Defaults to None.

    Example:
        >>>model = UBPPhase3(V=V, y=y, batch_size=32, missing_mask=missing_mask, output_shape=y_train.shape[1], n_components=3, weights_initializer="glorot_normal", hidden_layer_sizes="midpoint", num_hidden_layers=1, hidden_activation="elu", l1_penalty=1e-6, l2_penalty=1e-6, num_classes=3, phase=3)
        >>>
        >>>model.compile(optimizer=optimizer, loss=loss_func, metrics=[my_metrics], run_eagerly=True)
        >>>
        >>>history = model.fit(X, y, batch_size=batch_size, epochs=epochs, callbacks=[MyCallback()], validation_split=validation_split, shuffle=False)

    Raises:
        TypeError: V, y, missing_mask, output_shape must not be NoneType.
        ValueError: Maximum of 5 hidden layers.
    """

    def __init__(
        self,
        V=None,
        y=None,
        batch_size=32,
        missing_mask=None,
        output_shape=None,
        n_components=3,
        weights_initializer="glorot_normal",
        hidden_layer_sizes="midpoint",
        num_hidden_layers=1,
        hidden_activation="elu",
        dropout_rate=0.2,
        num_classes=3,
        phase=3,
        sample_weight=None,
    ):
        super(UBPPhase3, self).__init__()

        nn = NeuralNetworkMethods()
        self.nn = nn

        if V is None:
            self._V = nn.init_weights(y.shape[0], n_components)
        elif isinstance(V, dict):
            self._V = V[n_components]
        else:
            self._V = V

        self._y = y

        hidden_layer_sizes = nn.validate_hidden_layers(
            hidden_layer_sizes, num_hidden_layers
        )

        hidden_layer_sizes = nn.get_hidden_layer_sizes(
            y.shape[1], self._V.shape[1], hidden_layer_sizes
        )

        nn.validate_model_inputs(y, missing_mask, output_shape)

        self._missing_mask = missing_mask
        self.weights_initializer = weights_initializer
        self.phase = phase
        self.dropout_rate = dropout_rate
        self._sample_weight = sample_weight

        ### NOTE: I tried using just _V as the input to be refined, but it
        # wasn't getting updated. So I copy it here and it works.
        # V_latent is refined during train_step.
        self.V_latent_ = self._V.copy()

        # Initialize parameters used during train_step.
        self._batch_idx = 0
        self._batch_size = batch_size
        self.n_components = n_components

        # No regularization in phase 3.
        kernel_regularizer = None
        self.kernel_regularizer = kernel_regularizer
        kernel_initializer = None

        if hidden_activation.lower() == "leaky_relu":
            activation = LeakyReLU(alpha=0.01)

        elif hidden_activation.lower() == "prelu":
            activation = PReLU()

        elif hidden_activation.lower() == "selu":
            activation = "selu"
            kernel_initializer = "lecun_normal"

        else:
            activation = hidden_activation

        if num_hidden_layers > 5:
            raise ValueError(
                f"The maximum number of hidden layers is 5, but got "
                f"{num_hidden_layers}"
            )

        self.dense2 = None
        self.dense3 = None
        self.dense4 = None
        self.dense5 = None

        # Construct multi-layer perceptron.
        # Add hidden layers dynamically.
        self.dense1 = Dense(
            hidden_layer_sizes[0],
            input_shape=(n_components,),
            activation=activation,
            kernel_initializer=kernel_initializer,
        )

        if num_hidden_layers >= 2:
            self.dense2 = Dense(
                hidden_layer_sizes[1],
                activation=activation,
                kernel_initializer=kernel_initializer,
            )

        if num_hidden_layers >= 3:
            self.dense3 = Dense(
                hidden_layer_sizes[2],
                activation=activation,
                kernel_initializer=kernel_initializer,
            )

        if num_hidden_layers >= 4:
            self.dense4 = Dense(
                hidden_layer_sizes[3],
                activation=activation,
                kernel_initializer=kernel_initializer,
            )

        if num_hidden_layers == 5:
            self.dense5 = Dense(
                hidden_layer_sizes[4],
                activation=activation,
                kernel_initializer=kernel_initializer,
            )

        self.output1 = Dense(
            output_shape * num_classes,
            kernel_initializer=kernel_initializer,
        )

        self.rshp = Reshape((output_shape, num_classes))

        self.dropout_layer = Dropout(rate=dropout_rate)

    def call(self, inputs, training=None):
        x = self.dense1(inputs)
        if self.dense2 is not None:
            x = self.dense2(x)
        if self.dense3 is not None:
            x = self.dense3(x)
        if self.dense4 is not None:
            x = self.dense4(x)
        if self.dense5 is not None:
            x = self.dense5(x)

        x = self.output1(x)
        return self.rshp(x)

    def model(self):
        x = tf.keras.Input(shape=(self.n_components,))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def set_model_outputs(self):
        x = tf.keras.Input(shape=(self.n_components,))
        model = tf.keras.Model(inputs=[x], outputs=self.call(x))
        self.outputs = model.outputs

    def train_step(self, data):
        """Train step function. Parameters are set in the UBPCallbacks callback"""
        y = self._y

        (
            v,
            y_true,
            sample_weight,
            missing_mask,
            batch_start,
            batch_end,
        ) = self.nn.prepare_training_batches(
            self.V_latent_,
            y,
            self._batch_size,
            self._batch_idx,
            True,
            self.n_components,
            self._sample_weight,
            self._missing_mask,
        )

        src = [v]

        if sample_weight is not None:
            sample_weight_masked = tf.convert_to_tensor(
                sample_weight[~missing_mask], dtype=tf.float32
            )
        else:
            sample_weight_masked = None

        y_true_masked = tf.boolean_mask(
            tf.convert_to_tensor(y_true, dtype=tf.float32),
            tf.reduce_any(tf.not_equal(y_true, -1), axis=2),
        )

        # NOTE: Earlier model architectures incorrectly
        # applied one gradient to all the variables, including
        # the weights and v. Here we apply them separately, per
        # the UBP manuscript.
        with tf.GradientTape(persistent=True) as tape:
            # Forward pass. Watch input tensor v.
            tape.watch(v)
            y_pred = self(v, training=True)
            y_pred_masked = tf.boolean_mask(
                y_pred, tf.reduce_any(tf.not_equal(y_true, -1), axis=2)
            )
            ### NOTE: If you get the error, "'tuple' object has no attribute
            ### 'rank'", then convert y_true to a tensor object."
            loss = self.compiled_loss(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
                regularization_losses=self.losses,
            )

        # Refine the watched variables with
        # gradient descent backpropagation
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(
            zip(gradients, self.trainable_variables)
        )

        # Apply separate gradients to v.
        vgrad = tape.gradient(loss, src)
        self.optimizer.apply_gradients(zip(vgrad, src))

        del tape

        ### NOTE: If you get the error, "'tuple' object has no attribute
        ### 'rank', then convert y_true to a tensor object."
        self.compiled_metrics.update_state(
            y_true_masked,
            y_pred_masked,
            sample_weight=sample_weight_masked,
        )

        # NOTE: run_eagerly must be set to True in the compile() method for this
        # to work. Otherwise it can't convert a Tensor object to a numpy array.
        # There is really no other way to set v back to V_latent_ in graph
        # mode as far as I know. eager execution is slower, so it would be nice
        # to find a way to do this without converting to numpy.
        self.V_latent_[batch_start:batch_end, :] = v.numpy()

        # history object that gets returned from fit().
        return {m.name: m.result() for m in self.metrics}

    @property
    def V_latent(self):
        """Randomly initialized input variable that gets refined during training.
        :noindex:
        """
        return self.V_latent_

    @property
    def batch_size(self):
        """Batch (=step) size per epoch.
        :noindex:
        """
        return self._batch_size

    @property
    def batch_idx(self):
        """Current batch (=step) index.
        :noindex:
        """
        return self._batch_idx

    @property
    def y(self):
        """Full input dataset y.
        :noindex:
        """
        return self._y

    @property
    def missing_mask(self):
        """Missing mask of shape (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._missing_mask

    @property
    def sample_weight(self):
        """Sample weights of shpe (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._sample_weight

    @V_latent.setter
    def V_latent(self, value):
        """Set randomly initialized input variable. Refined during training.
        :noindex:
        """
        self.V_latent_ = value

    @batch_size.setter
    def batch_size(self, value):
        """Set batch_size parameter.
        :noindex:
        """
        self._batch_size = int(value)

    @batch_idx.setter
    def batch_idx(self, value):
        """Set current batch (=step) index.
        :noindex:
        """
        self._batch_idx = int(value)

    @y.setter
    def y(self, value):
        """Set y after each epoch.
        :noindex:
        """
        self._y = value

    @missing_mask.setter
    def missing_mask(self, value):
        """Set missing_mask after each epoch.
        :noindex:
        """
        self._missing_mask = value

    @sample_weight.setter
    def sample_weight(self, value):
        """Set sample_weight after each epoch.
        :noindex:
        """
        self._sample_weight = value



import logging
import os
import sys
import warnings
import math

# Import tensorflow with reduced warnings.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

import numpy as np
import pandas as pd
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns of GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated

from tensorflow.keras.layers import (
    Dropout,
    Dense,
    Reshape,
    Activation,
    Flatten,
    BatchNormalization,
    LeakyReLU,
    PReLU,
)

from tensorflow.keras.regularizers import l1_l2

# Custom Modules
try:
    from ...neural_network_methods import NeuralNetworkMethods
except (ModuleNotFoundError, ValueError, ImportError):
    from impute.unsupervised.neural_network_methods import NeuralNetworkMethods


class SoftOrdering1DCNN(tf.keras.Model):
    def __init__(
        self,
        y=None,
        output_shape=None,
        weights_initializer="glorot_normal",
        hidden_layer_sizes="midpoint",
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=1e-6,
        l2_penalty=1e-6,
        dropout_rate=0.2,
        num_classes=4,
        sample_weight=None,
        batch_size=32,
        missing_mask=None,
        activation=None,
        channel_increase_rate=2,
        initial_hidden_size=2048,
        num_groups=256,
    ):
        super(SoftOrdering1DCNN, self).__init__()

        self._y = y
        self._missing_mask = missing_mask
        self._sample_weight = sample_weight
        self._batch_idx = 0
        self._batch_size = batch_size
        self.output_activation = activation
        self.sample_weight = sample_weight

        self.nn_ = NeuralNetworkMethods()
        self.binary_accuracy = self.nn_.make_masked_binary_accuracy(
            is_vae=True
        )

        self.total_loss_tracker = tf.keras.metrics.Mean(name="loss")
        self.reconstruction_loss_tracker = tf.keras.metrics.Mean(
            name="reconstruction_loss"
        )
        self.kl_loss_tracker = tf.keras.metrics.Mean(name="kl_loss")
        self.accuracy_tracker = tf.keras.metrics.Mean(name="accuracy")

        # y_train[1] dimension.
        self.n_features = output_shape * num_classes

        self.weights_initializer = weights_initializer
        self.hidden_layer_sizes = hidden_layer_sizes
        self.num_hidden_layers = num_hidden_layers
        self.hidden_activation = hidden_activation
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.dropout_rate = dropout_rate
        self.num_classes = num_classes
        self.channel_increase_rate = channel_increase_rate
        self.initial_hidden_size = initial_hidden_size
        self.channel_size1 = num_groups
        self.channel_size2 = num_groups * 2
        self.channel_size3 = num_groups * 2

        nn = NeuralNetworkMethods()

        # hidden_layer_sizes = nn.validate_hidden_layers(
        #     self.hidden_layer_sizes, self.num_hidden_layers
        # )

        # hidden_layer_sizes = nn.get_hidden_layer_sizes(
        #     self.n_features, self.n_components, hidden_layer_sizes, vae=True
        # )

        # hidden_layer_sizes = [h * self.num_classes for h in hidden_layer_sizes]

        if self.l1_penalty == 0.0 and self.l2_penalty == 0.0:
            kernel_regularizer = None
        else:
            kernel_regularizer = l1_l2(self.l1_penalty, self.l2_penalty)

        kernel_initializer = self.weights_initializer

        if self.hidden_activation.lower() == "leaky_relu":
            activation = LeakyReLU(alpha=0.01)

        elif self.hidden_activation.lower() == "prelu":
            activation = PReLU()

        elif self.hidden_activation.lower() == "selu":
            activation = "selu"
            kernel_initializer = "lecun_normal"

        else:
            activation = self.hidden_activation

        if num_hidden_layers > 5:
            raise ValueError(
                f"The maximum number of hidden layers is 5, but got "
                f"{num_hidden_layers}"
            )

        hidden_size = initial_hidden_size
        if self.n_features >= hidden_size:
            scaling_factor = int(math.ceil(self.n_features / hidden_size)) * 2
            hidden_size *= num_groups * int(
                math.ceil((scaling_factor / num_groups))
            )
        else:
            # If hidden_size is close in number to n_features
            if abs(hidden_size - self.n_features) <= (hidden_size // 2):
                hidden_size *= 2

        # Model adapted from: https://medium.com/spikelab/convolutional-neural-networks-on-tabular-datasets-part-1-4abdd67795b6

        signal_size1 = hidden_size // num_groups
        signal_size2 = signal_size1 // 2
        signal_size3 = signal_size1 // 4 * self.channel_size3

        self.signal_size1 = signal_size1
        self.signal_size2 = signal_size2
        self.signal_size3 = signal_size3

        self.batch_norm1 = BatchNormalization()
        self.dropout1 = Dropout(self.dropout_rate)
        self.dense1 = Dense(
            hidden_size,
            input_shape=(self.n_features,),
            activation=hidden_activation,
            kernel_initializer=kernel_initializer,
        )

        self.rshp = Reshape((num_groups, signal_size1))

        self.batch_norm_c1 = BatchNormalization()
        self.conv1 = tf.keras.layers.Conv1D(
            self.channel_size1 * self.channel_increase_rate,
            kernel_size=5,
            stride=1,
            padding=2,
            groups=signal_size1,
            kernel_initializer=kernel_initializer,
            activation=hidden_activation,
        )

        self.avg_po_c1 = tf.keras.layers.AveragePooling1D(
            pool_size=4, padding="valid"
        )

        self.batch_norm_c2 = BatchNormalization()
        self.dropout_c2 = Dropout(self.dropout_rate)
        self.conv2 = tf.keras.layers.Conv1D(
            self.channel_size2,
            kernel_size=3,
            stride=1,
            padding=1,
            kernel_initializer=kernel_initializer,
            activation=hidden_activation,
        )

        self.batch_norm_c3 = BatchNormalization()
        self.dropout_c3 = Dropout(self.dropout_rate)
        self.conv3 = tf.keras.layers.Conv1D(
            self.channel_size2,
            kernel_size=3,
            stride=1,
            padding=1,
            kernel_initializer=kernel_initializer,
            activation=hidden_activation,
        )

        self.batch_norm_c4 = BatchNormalization()
        self.dropout_c4 = Dropout(self.dropout_rate)
        self.conv4 = tf.keras.layers.Conv1D(
            self.channel_size2,
            kernel_size=5,
            stride=1,
            padding=2,
            groups=signal_size1,
            kernel_initializer=kernel_initializer,
            activation=None,
        )

        self.act_c4 = Activation(hidden_activation)

        self.max_po_c4 = tf.keras.layers.MaxPooling1D(
            pool_size=4, stride=2, padding=1
        )

        self.flatten = Flatten()

        self.batch_norm2 = BatchNormalization()
        self.dropout2 = Dropout(self.dropout_rate)
        self.dense2 = Dense(
            self.n_features, kernel_initializer=kernel_initializer
        )
        self.rshp2 = Reshape((output_shape, num_classes))
        self.act2 = Activation(activation)

    def call(self, inputs, training=None):
        """Call the model on a particular input.

        Args:
            input (tf.Tensor): Input tensor. Must be one-hot encoded.

        Returns:
            tf.Tensor: Output predictions. Will be one-hot encoded.
        """
        x = self.dense1(inputs)
        x = self.batch_norm1(x, training=training)
        x = self.dropout1(x, training=training)
        x = self.rshp(x)
        x = self.conv1(x)
        x = self.batch_norm_c1(x, training=training)
        x = self.avg_po_c1(x)
        x = self.conv2(x)
        x = self.batch_norm(x, training=training)
        x = self.dropout_c2(x, training=training)
        x_s = x
        x = self.conv3(x)
        x = self.batch_norm_c3(x, training=training)
        x = self.dropout(x, training=training)
        x = self.conv4(x)
        x = self.batch_norm_c4(x, training=training)
        x += x_s
        x = self.act_c4(x)
        x = self.max_po_c4(x)
        x = self.dropout1(x)
        x = self.rshp(x)
        x = self.batch_norm_c1(x)
        x = self.conv1(x)
        x = self.avg_po_c1(x)
        x = self.flatten(x)
        x = self.dense2(x)
        x = self.batch_norm2(x, training=training)
        x = self.dropout2(x, training=training)
        x = self.rshp2(x)
        return self.act2(x)

    def model(self):
        """Here so that mymodel.model().summary() can be called for debugging."""
        x = tf.keras.Input(shape=(self.n_features * self.num_classes,))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def set_model_outputs(self):
        x = tf.keras.Input(shape=(self.n_features * self.num_classes,))
        model = tf.keras.Model(inputs=[x], outputs=self.call(x))
        self.outputs = model.outputs

    @property
    def metrics(self):
        return [
            self.total_loss_tracker,
            self.reconstruction_loss_tracker,
            self.kl_loss_tracker,
            self.accuracy_tracker,
        ]

    @tf.function
    def train_step(self, data):
        # if isinstance(data, tuple):
        #     if len(data) == 2:
        #         x, y = data
        #         sample_weight = None
        #     else:
        #         x, y, sample_weight = data
        # else:
        #     raise TypeError("Target y must be supplied to fit for this model.")

        # Set in the UBPCallbacks() callback.
        y = self._y

        (
            y,
            y_true,
            sample_weight,
            missing_mask,
            batch_start,
            batch_end,
        ) = self.nn_.prepare_training_batches(
            y,
            y,
            self._batch_size,
            self._batch_idx,
            True,
            self.n_components,
            self._sample_weight,
            self._missing_mask,
            ubp=False,
        )

        if sample_weight is not None:
            sample_weight_masked = tf.convert_to_tensor(
                sample_weight[~missing_mask], dtype=tf.float32
            )
        else:
            sample_weight_masked = None

        y_true_masked = tf.boolean_mask(
            tf.convert_to_tensor(y_true, dtype=tf.float32),
            tf.reduce_any(tf.not_equal(y_true, -1), axis=2),
        )

        with tf.GradientTape() as tape:
            reconstruction = self(tf.convert_to_tensor(y), training=True)

            y_pred_masked = tf.boolean_mask(
                reconstruction, tf.reduce_any(tf.not_equal(y_true, -1), axis=2)
            )

            # Returns binary crossentropy loss.
            loss = self.compiled_loss(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
            )

        grads = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))
        self.total_loss_tracker.update_state(loss)

        ### NOTE: If you get the error, "'tuple' object has no attribute
        ### 'rank', then convert y_true to a tensor object."
        # self.compiled_metrics.update_state(
        self.accuracy_tracker.update_state(
            self.binary_accuracy(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
            )
        )

        return {
            "loss": self.total_loss_tracker.result(),
            "accuracy": self.accuracy_tracker.result(),
        }

    @tf.function
    def test_step(self, data):
        if isinstance(data, tuple):
            if len(data) == 2:
                x, y = data
                sample_weight = None
            else:
                x, y, sample_weight = data
        else:
            raise TypeError("Target y must be supplied to fit in this model.")

        if sample_weight is not None:
            sample_weight_masked = tf.boolean_mask(
                tf.convert_to_tensor(sample_weight),
                tf.reduce_any(tf.not_equal(y, -1), axis=2),
            )
        else:
            sample_weight_masked = None

        reconstruction, z_mean, z_log_var, z = self(x, training=False)
        reconstruction_loss = self.compiled_loss(
            y,
            reconstruction,
            sample_weight=sample_weight_masked,
        )

        # Includes KL Divergence Loss.
        regularization_loss = sum(self.losses)

        total_loss = reconstruction_loss + regularization_loss

        self.accuracy_tracker.update_state(
            self.cateogrical_accuracy(
                y,
                reconstruction,
                sample_weight=sample_weight_masked,
            )
        )

        self.total_loss_tracker.update_state(total_loss)
        self.reconstruction_loss_tracker.update_state(reconstruction_loss)
        self.kl_loss_tracker.update_state(regularization_loss)

        return {
            "loss": self.total_loss_tracker.result(),
            "reconstruction_loss": self.reconstruction_loss_tracker.result(),
            "kl_loss": self.kl_loss_tracker.result(),
            "accuracy": self.accuracy_tracker.result(),
        }

    @property
    def batch_size(self):
        """Batch (=step) size per epoch."""
        return self._batch_size

    @property
    def batch_idx(self):
        """Current batch (=step) index."""
        return self._batch_idx

    @property
    def y(self):
        return self._y

    @property
    def missing_mask(self):
        return self._missing_mask

    @property
    def sample_weight(self):
        return self._sample_weight

    @batch_size.setter
    def batch_size(self, value):
        """Set batch_size parameter."""
        self._batch_size = int(value)

    @batch_idx.setter
    def batch_idx(self, value):
        """Set current batch (=step) index."""
        self._batch_idx = int(value)

    @y.setter
    def y(self, value):
        """Set y after each epoch."""
        self._y = value

    @missing_mask.setter
    def missing_mask(self, value):
        """Set y after each epoch."""
        self._missing_mask = value

    @sample_weight.setter
    def sample_weight(self, value):
        self._sample_weight = value



import logging
import os
import sys
import warnings

# Import tensorflow with reduced warnings.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

# noinspection PyPackageRequirements
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns off GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated

from tensorflow.keras.layers import (
    Dropout,
    Dense,
    Reshape,
    LeakyReLU,
    PReLU,
)

from tensorflow.keras.regularizers import l1_l2

# Custom Modules
try:
    from ..neural_network_methods import NeuralNetworkMethods
except (ModuleNotFoundError, ValueError, ImportError):
    from impute.unsupervised.neural_network_methods import NeuralNetworkMethods


class NLPCAModel(tf.keras.Model):
    """NLPCA model to train and use to predict imputations.

    NLPCAModel subclasses the tf.keras.Model and overrides the train_step function, which does training and evaluation for each batch in each epoch.

    Args:
        V (numpy.ndarray(float)): V should have been randomly initialized and will be used as the input data that gets refined during training. Defaults to None.

        y (numpy.ndarray): Target values to predict. Actual input data. Defaults to None.

        batch_size (int, optional): Batch size per epoch. Defaults to 32.

        missing_mask (numpy.ndarray): Missing data mask for y. Defaults to None.

        output_shape (int): Output units for n_features dimension. Output will be of shape (batch_size, n_features). Defaults to None.

        n_components (int, optional): Number of features in input V to use. Defaults to 3.

        weights_initializer (str, optional): Kernel initializer to use for initializing model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (List[int], optional): Output units for each hidden layer. List should be of same length as the number of hidden layers. Defaults to "midpoint".

        num_hidden_layers (int, optional): Number of hidden layers to use. Defaults to 1.

        hidden_activation (str, optional): Activation function to use for hidden layers. Defaults to "elu".

        l1_penalty (float, optional): L1 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate during training to reduce overfitting. Must be a float between 0 and 1. Defaults to 0.2.

        num_classes (int, optional): Number of classes in output. Corresponds to the 3rd dimension of the output shape (batch_size, n_features, num_classes). Defaults to 1.

        phase (NoneType): Here for compatibility with UBP.

        sample_weight (numpy.ndarray, optional): 2D sample weights of shape (n_samples, n_features). Should have values for each class weighted. Defaults to None.

    Example:
        >>>model = NLPCAModel(V=V, y=y, batch_size=32, missing_mask=missing_mask, output_shape, n_components, weights_initializer, hidden_layer_sizes, num_hidden_layers, hidden_activation, l1_penalty, l2_penalty, dropout_rate, num_classes=3)
        >>>
        >>>model.compile(optimizer=optimizer, loss=loss_func, metrics=[my_metrics], run_eagerly=True)
        >>>
        >>>history = model.fit(X, y, batch_size=batch_size, epochs=epochs, callbacks=[MyCallback()], validation_split=validation_split, shuffle=False)

    Raises:
        TypeError: V, y, missing_mask, output_shape must not be NoneType.
        ValueError: Maximum of 5 hidden layers.

    """

    def __init__(
        self,
        V=None,
        y=None,
        batch_size=32,
        missing_mask=None,
        output_shape=None,
        n_components=3,
        weights_initializer="glorot_normal",
        hidden_layer_sizes="midpoint",
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        num_classes=3,
        phase=None,
        sample_weight=None,
    ):
        super(NLPCAModel, self).__init__()

        nn = NeuralNetworkMethods()
        self.nn = nn

        if V is None:
            self._V = nn.init_weights(y.shape[0], n_components)
        elif isinstance(V, dict):
            self._V = V[n_components]
        else:
            self._V = V

        self._y = y

        hidden_layer_sizes = nn.validate_hidden_layers(
            hidden_layer_sizes, num_hidden_layers
        )

        hidden_layer_sizes = nn.get_hidden_layer_sizes(
            y.shape[1], self._V.shape[1], hidden_layer_sizes
        )

        nn.validate_model_inputs(y, missing_mask, output_shape)

        self._missing_mask = missing_mask
        self.weights_initializer = weights_initializer
        self.phase = phase
        self.dropout_rate = dropout_rate
        self._sample_weight = sample_weight

        ### NOTE: I tried using just _V as the input to be refined, but it
        # wasn't getting updated. So I copy it here and it works.
        # V_latent is refined during train_step.
        self.V_latent_ = self._V.copy()

        # Initialize parameters used during train_step.
        self._batch_idx = 0
        self._batch_size = batch_size
        self.n_components = n_components

        if l1_penalty == 0.0 and l2_penalty == 0.0:
            kernel_regularizer = None
        else:
            kernel_regularizer = l1_l2(l1_penalty, l2_penalty)

        self.kernel_regularizer = kernel_regularizer
        kernel_initializer = weights_initializer

        if hidden_activation.lower() == "leaky_relu":
            activation = LeakyReLU(alpha=0.01)

        elif hidden_activation.lower() == "prelu":
            activation = PReLU()

        elif hidden_activation.lower() == "selu":
            activation = "selu"
            kernel_initializer = "lecun_normal"

        else:
            activation = hidden_activation

        if num_hidden_layers > 5:
            raise ValueError(
                f"The maximum number of hidden layers is 5, but got "
                f"{num_hidden_layers}"
            )

        self.dense2 = None
        self.dense3 = None
        self.dense4 = None
        self.dense5 = None

        # Construct multi-layer perceptron.
        # Add hidden layers dynamically.
        self.dense1 = Dense(
            hidden_layer_sizes[0],
            input_shape=(n_components,),
            activation=activation,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
        )

        if num_hidden_layers >= 2:
            self.dense2 = Dense(
                hidden_layer_sizes[1],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers >= 3:
            self.dense3 = Dense(
                hidden_layer_sizes[2],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers >= 4:
            self.dense4 = Dense(
                hidden_layer_sizes[3],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        if num_hidden_layers == 5:
            self.dense5 = Dense(
                hidden_layer_sizes[4],
                activation=activation,
                kernel_initializer=kernel_initializer,
                kernel_regularizer=kernel_regularizer,
            )

        self.output1 = Dense(
            output_shape * num_classes,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
        )

        self.rshp = Reshape((output_shape, num_classes))

        self.dropout_layer = Dropout(rate=dropout_rate)

    def call(self, inputs, training=None):
        x = self.dense1(inputs)
        x = self.dropout_layer(x, training=training)
        if self.dense2 is not None:
            x = self.dense2(x)
            x = self.dropout_layer(x, training=training)
        if self.dense3 is not None:
            x = self.dense3(x)
            x = self.dropout_layer(x, training=training)
        if self.dense4 is not None:
            x = self.dense4(x)
            x = self.dropout_layer(x, training=training)
        if self.dense5 is not None:
            x = self.dense5(x)
            x = self.dropout_layer(x, training=training)

        x = self.output1(x)
        return self.rshp(x)

    def model(self):
        x = tf.keras.Input(shape=(self.n_components,))
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

    def set_model_outputs(self):
        x = tf.keras.Input(shape=(self.n_components,))
        model = tf.keras.Model(inputs=[x], outputs=self.call(x))
        self.outputs = model.outputs

    def train_step(self, data):
        """Train step function. Parameters are set in UBPCallbacks callback."""
        y = self._y

        (
            v,
            y_true,
            sample_weight,
            missing_mask,
            batch_start,
            batch_end,
        ) = self.nn.prepare_training_batches(
            self.V_latent_,
            y,
            self._batch_size,
            self._batch_idx,
            True,
            self.n_components,
            self._sample_weight,
            self._missing_mask,
        )

        src = [v]

        if sample_weight is not None:
            sample_weight_masked = tf.convert_to_tensor(
                sample_weight[~missing_mask], dtype=tf.float32
            )
        else:
            sample_weight_masked = None

        y_true_masked = tf.boolean_mask(
            tf.convert_to_tensor(y_true, dtype=tf.float32),
            tf.reduce_any(tf.not_equal(y_true, -1), axis=2),
        )

        # NOTE: Earlier model architectures incorrectly
        # applied one gradient to all the variables, including
        # the weights and v. Here we apply them separately, per
        # the UBP manuscript.
        with tf.GradientTape(persistent=True) as tape:
            # Forward pass. Watch input tensor v.
            tape.watch(v)
            y_pred = self(v, training=True)
            y_pred_masked = tf.boolean_mask(
                y_pred, tf.reduce_any(tf.not_equal(y_true, -1), axis=2)
            )
            ### NOTE: If you get the error, "'tuple' object has no attribute
            ### 'rank'", then convert y_true to a tensor object."
            loss = self.compiled_loss(
                y_true_masked,
                y_pred_masked,
                sample_weight=sample_weight_masked,
                regularization_losses=self.losses,
            )

        # Refine the watched variables with
        # gradient descent backpropagation
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(
            zip(gradients, self.trainable_variables)
        )

        # Apply separate gradients to v.
        vgrad = tape.gradient(loss, src)
        self.optimizer.apply_gradients(zip(vgrad, src))

        del tape

        ### NOTE: If you get the error, "'tuple' object has no attribute
        ### 'rank', then convert y_true to a tensor object."
        self.compiled_metrics.update_state(
            y_true_masked,
            y_pred_masked,
            sample_weight=sample_weight_masked,
        )

        # NOTE: run_eagerly must be set to True in the compile() method for this
        # to work. Otherwise it can't convert a Tensor object to a numpy array.
        # There is really no other way to set v back to V_latent_ in graph
        # mode as far as I know. eager execution is slower, so it would be nice
        # to find a way to do this without converting to numpy.
        self.V_latent_[batch_start:batch_end, :] = v.numpy()

        # history object that gets returned from fit().
        return {m.name: m.result() for m in self.metrics}

    @property
    def V_latent(self):
        """Randomly initialized input that gets refined during training.
        :noindex:
        """
        return self.V_latent_

    @property
    def batch_size(self):
        """Batch (=step) size per epoch.
        :noindex:
        """
        return self._batch_size

    @property
    def batch_idx(self):
        """Current batch (=step) index.
        :noindex:
        """
        return self._batch_idx

    @property
    def y(self):
        """Full dataset y.
        :noindex:
        """
        return self._y

    @property
    def missing_mask(self):
        """Missing mask of shape (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._missing_mask

    @property
    def sample_weight(self):
        """Sample weights of shape (y.shape[0], y.shape[1])
        :noindex:
        """
        return self._sample_weight

    @V_latent.setter
    def V_latent(self, value):
        """Set randomly initialized input. Refined during training.
        :noindex:
        """
        self.V_latent_ = value

    @batch_size.setter
    def batch_size(self, value):
        """Set batch_size parameter.
        :noindex:
        """
        self._batch_size = int(value)

    @batch_idx.setter
    def batch_idx(self, value):
        """Set current batch (=step) index.
        :noindex:
        """
        self._batch_idx = int(value)

    @y.setter
    def y(self, value):
        """Set y after each epoch.
        :noindex:
        """
        self._y = value

    @missing_mask.setter
    def missing_mask(self, value):
        """Set missing_mask after each epoch.
        :noindex:
        """
        self._missing_mask = value

    @sample_weight.setter
    def sample_weight(self, value):
        """Set sample_weight after each epoch.
        :noindex:
        """
        self._sample_weight = value
